#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought util package component>
#------------------------------------------------------------------------------
"""
Copied from Python Cookbook.

"""

class RingBuffer:
    def __init__(self,size_max):
        self.max = size_max
        self.data = []
    def append(self,x):
        """append an element at the end of the buffer"""
        self.data.append(x)
        if len(self.data) == self.max:
            self.cur = 0
            self.__class__ = RingBufferFull
    def get(self):
        """ return a list of elements from the oldest to the newest"""
        return self.data


class RingBufferFull:
    def __init__(self,n):
        raise Exception("you should use RingBuffer")
    def append(self,x):
        self.data[self.cur]=x
        self.cur=(self.cur+1) % self.max
    def get(self):
        return self.data[self.cur:]+self.data[:self.cur]


# sample of use
"""x=RingBuffer(5)
x.append(1); x.append(2); x.append(3); x.append(4)
print x.__class__,x.get()
x.append(5)
print x.__class__,x.get()
x.append(6)
print x.data,x.get()
x.append(7); x.append(8); x.append(9); x.append(10)
print x.data,x.get()"""



#
#  Copyright (c) 2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Author: Senganal T.
#  Date:   Feb 2015
#

"""
Test cases for the Controller class.
"""

import nose

from traits.api import HasTraits, Instance, Str
from traitsui.api import Controller


class FooModel(HasTraits):
    my_str = Str('hallo')


class FooController(Controller):
    """ Test dialog that does nothing useful."""

    model = Instance(FooModel)

    def _model_default(self):
        return FooModel(my_str='meh')


def test_construction():
    # check default constructor.
    dialog = FooController()
    nose.tools.assert_is_not_none(dialog.model)
    nose.tools.assert_equal(dialog.model.my_str, 'meh')

    # check initialization when `model` is explcitly passed in.
    new_model = FooModel()
    dialog = FooController(model=new_model)
    nose.tools.assert_is(dialog.model, new_model)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/07/2004
#
#------------------------------------------------------------------------------
# Adding this statement for backwards compatibility (since editors.py was a
# file prior to version 3.0.3).

from __future__ import absolute_import

try:
    from .api import ArrayEditor
except ImportError:
    pass

from .api import (
    toolkit,
    BooleanEditor,
    ButtonEditor,
    CheckListEditor,
    CodeEditor,
    ColorEditor,
    CompoundEditor,
    CustomEditor,
    DateEditor,
    DefaultOverride,
    DirectoryEditor,
    DNDEditor,
    DropEditor,
    EnumEditor,
    FileEditor,
    FontEditor,
    KeyBindingEditor,
    ImageEditor,
    ImageEnumEditor,
    InstanceEditor,
    ListEditor,
    ListStrEditor,
    NullEditor,
    RangeEditor,
    RGBColorEditor,
    SetEditor,
    TextEditor,
    TableEditor,
    TimeEditor,
    TitleEditor,
    TreeEditor,
    TupleEditor,
    HistoryEditor,
    HTMLEditor,
    PopupEditor,
    ValueEditor,
    ShellEditor,
    ScrubberEditor,
    TabularEditor,
    ProgressEditor,
    SearchEditor)



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The abstract interface for all pyface top-level windows. """


# Enthought library imports.
from traits.api import Event, Tuple, Unicode

# Local imports.
from pyface.constant import NO
from pyface.key_pressed_event import KeyPressedEvent
from pyface.i_widget import IWidget


class IWindow(IWidget):
    """ The abstract interface for all pyface top-level windows.

    A pyface top-level window has no visual representation until it is opened
    (ie. its 'control' trait will be None until it is opened).
    """

    #### 'IWindow' interface ##################################################

    #: The position of the window.
    position = Tuple

    #: The size of the window.
    size = Tuple

    #: The window title.
    title = Unicode

    #### Events #####

    #: The window has been activated.
    activated = Event

    #: The window has been closed.
    closed =  Event

    #: The window is about to be closed.
    closing =  Event

    #: The window has been deactivated.
    deactivated = Event

    #: A key was pressed while the window had focus.
    # FIXME v3: This smells of a hack. What's so special about key presses?
    # FIXME v3: Unicode
    key_pressed = Event(KeyPressedEvent)

    #: The window has been opened.
    opened = Event

    #: The window is about to open.
    opening = Event

    ###########################################################################
    # 'IWindow' interface.
    ###########################################################################

    def open(self):
        """ Opens the window. """

    def close(self):
        """ Closes the window. """

    def activate(self):
        """ Activates the window. """

    def show(self, visible):
        """ Show or hide the window.

        Parameter
        ---------
        visible : bool
            Visible should be ``True`` if the window should be shown.
        """

    def confirm(self, message, title=None, cancel=False, default=NO):
        """ Convenience method to show a confirmation dialog.

        Parameters
        ----------
        message : str
            The text of the message to display.
        title : str
            The text of the dialog title.
        cancel : bool
            ``True`` if the dialog should contain a Cancel button.
        default : NO, YES or CANCEL
            Which button should be the default button.
        """

    def information(
            self, message, title='Information', detail='', informative=''):
        """ Convenience method to show an information message dialog.

        Parameters
        ----------
        message : str
            The text of the message to display.
        title : str
            The text of the dialog title.
        detail : str
            Further details about the message.
        informative : str
            Explanatory text to display along with the message.

        """

    def warning(self, message, title='Warning', detail='', informative=''):
        """ Convenience method to show a warning message dialog.

        Parameters
        ----------
        message : str
            The text of the message to display.
        title : str
            The text of the dialog title.
        detail : str
            Further details about the message.
        informative : str
            Explanatory text to display along with the message.

        """

    def error(self, message, title='Error', detail='', informative=''):
        """ Convenience method to show an error message dialog.

        Parameters
        ----------
        message : str
            The text of the message to display.
        title : str
            The text of the dialog title.
        detail : str
            Further details about the message.
        informative : str
            Explanatory text to display along with the message.

        """

    ###########################################################################
    # Protected 'IWindow' interface.
    ###########################################################################

    def _add_event_listeners(self):
        """ Adds any event listeners required by the window. """


class MWindow(object):
    """ The mixin class that contains common code for toolkit specific
    implementations of the IWindow interface.

    Implements: close(), confirm(), open()
    Reimplements: _create()
    """

    ###########################################################################
    # 'IWindow' interface.
    ###########################################################################

    def open(self):
        """ Opens the window. """

        # Trait notification.
        self.opening = self

        if self.control is None:
            self._create()

        self.show(True)

        # Trait notification.
        self.opened = self

    def close(self):
        """ Closes the window. """

        if self.control is not None:
            # Trait notification.
            self.closing = self

            # Cleanup the toolkit-specific control.
            self.destroy()

            # Trait notification.
            self.closed = self

    def confirm(self, message, title=None, cancel=False, default=NO):
        """ Convenience method to show a confirmation dialog.

        Parameters
        ----------
        message : str
            The text of the message to display.
        title : str
            The text of the dialog title.
        cancel : bool
            ``True`` if the dialog should contain a Cancel button.
        default : NO, YES or CANCEL
            Which button should be the default button.
        """
        from confirmation_dialog import confirm

        return confirm(self.control, message, title, cancel, default)

    def information(
            self, message, title='Information', detail='', informative=''):
        """ Convenience method to show an information message dialog.

        Parameters
        ----------
        message : str
            The text of the message to display.
        title : str
            The text of the dialog title.
        detail : str
            Further details about the message.
        informative : str
            Explanatory text to display along with the message.

        """
        from message_dialog import information

        information(self.control, message, title, detail, informative)

    def warning(self, message, title='Warning', detail='', informative=''):
        """ Convenience method to show a warning message dialog.

        Parameters
        ----------
        message : str
            The text of the message to display.
        title : str
            The text of the dialog title.
        detail : str
            Further details about the message.
        informative : str
            Explanatory text to display along with the message.

        """
        from message_dialog import warning

        warning(self.control, message, title, detail, informative)

    def error(self, message, title='Error', detail='', informative=''):
        """ Convenience method to show an error message dialog.

        Parameters
        ----------
        message : str
            The text of the message to display.
        title : str
            The text of the dialog title.
        detail : str
            Further details about the message.
        informative : str
            Explanatory text to display along with the message.

        """
        from message_dialog import error

        error(self.control, message, title, detail, informative)

    ###########################################################################
    # Protected 'IWidget' interface.
    ###########################################################################

    def _create(self):
        """ Creates the window's widget hierarchy. """

        # Create the toolkit-specific control.
        super(MWindow, self)._create()

        # Wire up event any event listeners required by the window.
        self._add_event_listeners()



#!/usr/bin/env python

"""
Tests for writing and reading UGRID compliant netCDF.

"""

from __future__ import (absolute_import, division, print_function)

import os
import numpy as np

from pyugrid.ugrid import UGrid, UVar

from .utilities import chdir, two_triangles


test_files = os.path.join(os.path.dirname(__file__), 'files')


def test_with_faces():
    """
    Test with faces, edges, but no `face_coordinates` or `edge_coordinates`.

    """

    expected = two_triangles()

    fname = '2_triangles.nc'
    with chdir(test_files):
        expected.save_as_netcdf(fname)
        grid = UGrid.from_ncfile(fname)
        os.remove(fname)

    assert np.array_equal(expected.nodes, grid.nodes)
    assert np.array_equal(expected.faces, grid.faces)
    assert np.array_equal(expected.edges, grid.edges)


def test_without_faces():
    expected = two_triangles()
    del expected.faces
    assert expected.faces is None

    fname = '2_triangles.nc'
    with chdir(test_files):
        expected.save_as_netcdf(fname)
        grid = UGrid.from_ncfile(fname)
        os.remove(fname)

    assert grid.faces is None
    assert np.array_equal(expected.faces, grid.faces)
    assert np.array_equal(expected.edges, grid.edges)


def test_with_just_nodes_and_depths():
    expected = two_triangles()
    del expected.faces
    del expected.edges

    depth = UVar('depth',
                 'node',
                 np.array([1.0, 2.0, 3.0, 4.0]),
                 {'units': 'm',
                  'positive': 'down',
                  'standard_name': 'sea_floor_depth_below_geoid'})
    expected.add_data(depth)

    fname = '2_triangles_depth.nc'
    with chdir(test_files):
        expected.save_as_netcdf(fname)
        grid = UGrid.from_ncfile(fname, load_data=True)
        os.remove(fname)

    assert grid.faces is None
    assert grid.edges is None
    assert np.array_equal(expected.nodes, grid.nodes)

    assert np.array_equal(expected.data['depth'].data, grid.data['depth'].data)
    assert expected.data['depth'].attributes == grid.data['depth'].attributes


if __name__ == "__main__":
    test_with_faces()
    test_without_faces()



"""GLSVG library for SVG rendering in PyOpenGL.

Example usage:
    $ import glsvg
    $ my_svg = glsvg.SVG('filename.svg')
    $ my_svg.draw(100, 200, angle=15)

"""

import OpenGL.GL as gl

try:
    import xml.etree.ElementTree
    from xml.etree.cElementTree import parse
except:
    import elementtree.ElementTree
    from elementtree.ElementTree import parse

import re
import math
import string
import traceback

from svg_constants import *

from glutils import *
from vector_math import *
from svg_parser_utils import parse_color, parse_float, parse_style, parse_list
from gradient import *

from svg_path import SVGPath, SVGGroup, SVGDefs, SVGUse, SVGMarker, SVGContainer
from svg_pattern import *
import graphics

from render_target import CanvasManager

class SVGConfig:
    """Configuration for how to render SVG objects, such as
    the amount of detail allowed for bezier curves and availability of the stencil buffer"""

    def __init__(self):

        self.use_fxaa = True

        #: The number of stencil bits available
        self.stencil_bits = gl.glGetInteger(gl.GL_STENCIL_BITS)

        #: Whether or not framebuffer objects are allowed
        self.has_framebuffer_objects = True

        #: Whether or not stencilling is allowed
        self.allow_stencil = self.stencil_bits > 0

        #: The number of line segments into which to subdivide Bezier splines.
        self.bezier_points = BEZIER_POINTS

        #: The number of line segments into which to subdivide circular and elliptic arcs.
        self.circle_points = CIRCLE_POINTS

        #: The minimum distance at which neighboring points are merged
        self.tolerance = TOLERANCE

    def super_detailed(self):
        """Returns a much more detailed copy of this config, for patterns"""

        cfg = SVGConfig()
        cfg.bezier_points *= 10
        cfg.circle_points *= 10
        cfg.tolerance /= 100
        return cfg

    def __repr__(self):
        return "<SVGConfig stencil_bits={0} fbo={1} circle_points={2} bezier_points={3}>".format(
            self.stencil_bits,
            self.has_framebuffer_objects,
            self.circle_points,
            self.bezier_points
        )


class SVGDoc(SVGContainer):
    """
    An SVG image document.

    Users should instantiate this object once for each SVG file they wish to
    render.

    """
    def __init__(self, filename_or_element, parent=None, anchor_x=0, anchor_y=0, config=None):
        """Creates an SVG document from a .svg or .svgz file.

        Args:
            `filename`: str
                The name of the file to be loaded.
            `anchor_x`: float
                The horizontal anchor position for scaling and rotations. Defaults to 0. The symbolic
                values 'left', 'center' and 'right' are also accepted.
            `anchor_y`: float
                The vertical anchor position for scaling and rotations. Defaults to 0. The symbolic
                values 'bottom', 'center' and 'top' are also accepted.
        """

        SVGContainer.__init__(self, parent)

        if not config:
            self.config = SVGConfig()
        else:
            self.config = config
        self._stencil_mask = 0

        #: Number of triangles in document
        self.n_tris = 0

        #: Number of lines in document
        self.n_lines = 0

        #: Map from id to path
        self.path_lookup = {}

        #: SVG paths
        self._paths = []

        #: Maps from pattern id to pattern
        self.patterns = {}

        #: Maps from id to marker def
        self.markers = {}

        #: Maps from id to path definition
        self.defs = {}

        #: Filename of original SVG file
        self.filename = filename_or_element if isinstance(filename_or_element, str) else None
        self._gradients = GradientContainer()

        if self.filename:
            if open(self.filename, 'rb').read(3) == '\x1f\x8b\x08':  # gzip magic numbers
                import gzip
                f = gzip.open(self.filename, 'rb')
            else:
                f = open(self.filename, 'rb')
            self.root = parse(f)._root
        else:
            self.root = filename_or_element

        self.parse_root(self.root)

        self._generate_disp_list()

        self.anchor_x = anchor_x
        self.anchor_y = anchor_y

    def parse_root(self, root):
        self._paths = []

        wm = root.get("width", '0')
        hm = root.get("height", '0')

        self.x = 0
        self.y = 0

        self.width = parse_float(wm)
        self.height = parse_float(hm)

        self.preserve_aspect_ratio = root.get('preserveAspectRatio', 'none')

        if self.root.get("viewBox"):
            x, y, w, h = (parse_float(x) for x in parse_list(root.get("viewBox")))
            self.x = x
            self.y = y
            self.height = h
            self.width = w

        self.opacity = 1.0
        for e in root.getchildren():
            try:
                self._parse_element(e)
            except Exception as ex:
                print 'Exception while parsing element', e
                raise

    def _is_path_tag(self, e):
        return (e.tag.endswith('path')
                or e.tag.endswith('rect')
                or e.tag.endswith('polyline') or e.tag.endswith('polygon')
                or e.tag.endswith('line')
                or e.tag.endswith('circle') or e.tag.endswith('ellipse'))

    def _parse_element(self, e, parent=None):
        renderable = None
        if self._is_path_tag(e):
            renderable = SVGPath(self, e, parent)
            if not parent:
                self._paths.append(renderable)

            if renderable.id:
                self.path_lookup[renderable.id] = renderable
        elif e.tag.endswith('}g') or e.tag == 'g':
            renderable = SVGGroup(self, e, parent)
            if renderable.id:
                self.path_lookup[renderable.id] = renderable
                self.defs[renderable.id] = renderable
            if not parent and not renderable.is_def:
                self._paths.append(renderable)
        elif e.tag.endswith('svg'):
            renderable = SVGDoc(e, parent)
            self._paths.append(renderable)
        elif e.tag.endswith('marker'):
            renderable = SVGMarker(self, e, parent)
        elif e.tag.endswith("text"):
            self._warn("Text tag not supported")
        elif e.tag.endswith('linearGradient'):
            self._gradients[e.get('id')] = LinearGradient(e, self)
        elif e.tag.endswith('radialGradient'):
            self._gradients[e.get('id')] = RadialGradient(e, self)
        elif e.tag.endswith('pattern'):
            renderable = SVGPattern(self, e, parent)
            self.patterns[e.get('id')] = renderable
        elif e.tag.endswith('defs'):
            renderable = SVGDefs(self, e, parent)
        elif e.tag.endswith('marker'):
            renderable = SVGMarker(self, e, parent)
        elif e.tag.endswith('use'):
            renderable = SVGUse(self, e, parent)
            self._paths.append(renderable)
        for c in e.getchildren():
            try:
                self._parse_element(c, renderable)
            except Exception, ex:
                print 'Exception while parsing element', c
                raise


    def get_path_ids(self):
        """Returns all the path ids"""
        return self.path_lookup.keys()

    def get_path_by_id(self, id):
        """Returns a path for the given id, or key error"""
        return self.path_lookup[id]

    def _register_pattern_part(self, pattern_id, pattern_svg_path):
        print "registering pattern"
        self.patterns[pattern_id].paths.append(pattern_svg_path)

    def _set_anchor_x(self, anchor_x):
        self._anchor_x = anchor_x
        if self._anchor_x == 'left':
            self._a_x = 0
        elif self._anchor_x == 'center':
            self._a_x = self.width * .5
        elif self._anchor_x == 'right':
            self._a_x = self.width
        else:
            self._a_x = self._anchor_x

    def _get_anchor_x(self):
        return self._anchor_x

    #: Where the document is anchored. Valid values are numerical, or 'left', 'right', 'center'
    anchor_x = property(_get_anchor_x, _set_anchor_x)

    def _set_anchor_y(self, anchor_y):
        self._anchor_y = anchor_y
        if self._anchor_y == 'bottom':
            self._a_y = 0
        elif self._anchor_y == 'center':
            self._a_y = self.height * .5
        elif self._anchor_y == 'top':
            self._a_y = self.height
        else:
            self._a_y = self.anchor_y

    def _get_anchor_y(self):
        return self._anchor_y

    #: Where the document is anchored. Valid values are numerical, or 'top', 'bottom', 'center'
    anchor_y = property(_get_anchor_y, _set_anchor_y)

    def _generate_disp_list(self):

        # prepare all the patterns
        self.prerender_patterns()

        # prepare all the predefined paths
        self.prerender_defs()

        with DisplayListGenerator() as display_list:
            self.disp_list = display_list
            self.render()

    def draw(self, x, y, z=0, angle=0, scale=1):
        """Draws the SVG to screen.

        Args:
            `x` : float
                The x-coordinate at which to draw.
            `y` : float
                The y-coordinate at which to draw.
            `z` : float
                The z-coordinate at which to draw. Defaults to 0. Note that z-ordering may not
                give expected results when transparency is used.
            `angle` : float
                The angle by which the image should be rotated (in degrees). Defaults to 0.
            `scale` : float
                The amount by which the image should be scaled, either as a float, or a tuple
                of two floats (xscale, yscale).

        """
        #CanvasManager.inst().update()
        #bg = CanvasManager.inst().get('BackgroundImage')

        with CurrentTransform():
            gl.glTranslatef(x, y, z)
            if angle:
                gl.glRotatef(angle, 0, 0, 1)
            if scale != 1:
                try:
                    gl.glScalef(scale[0], scale[1], 1)
                except TypeError:
                    gl.glScalef(scale, scale, 1)
            if self._a_x or self._a_y:
                gl.glTranslatef(-self._a_x, -self._a_y, 0)

            #with bg:
            gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
            self.disp_list()
        #bg.blit()

    def prerender_defs(self):
        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

        for d in self.defs.values():
            d.render()

    def prerender_patterns(self):
        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

        for pattern in self.patterns.values():
            pattern.render()

    def render(self):
        """Render the SVG file without any display lists or transforms. Use draw instead. """
        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

        graphics.clear_stats()
        #clear out stencils
        with Matrix.translation(self.x, self.y):
            for svg_path in self._paths:
                svg_path.render()

    def _warn(self, message):
        print "Warning: SVG Parser (%s) - %s" % (self.filename, message)



""" The menu bar manager for Envisage workbench windows. """


# Enthought library imports.
from pyface.action.api import MenuBarManager as BaseMenuBarManager
from traits.api import Instance

# Local imports.
from .action_controller import ActionController


class MenuBarManager(BaseMenuBarManager):
    """ The menu bar manager for Envisage workbench windows. """

    #### 'MenuBarManager' interface ###########################################

    # The workbench window that we are the menu bar manager for.
    window = Instance('pyface.workbench.api.WorkbenchWindow')

    ###########################################################################
    # 'MenuBarManager' interface.
    ###########################################################################

    def create_menu_bar(self, parent):
        """ Creates a menu bar representation of the manager. """

        # The controller handles the invocation of every action.
        controller = ActionController(window=self.window)

        menu_bar = super(MenuBarManager, self).create_menu_bar(
            parent, controller=controller
        )

        return menu_bar

#### EOF ######################################################################



__all__ = ["serve_fs"]

import SimpleHTTPServer
import SocketServer
from fs.path import pathjoin, dirname
from fs.errors import FSError
from time import mktime
from cStringIO import StringIO
import cgi
import urllib
import posixpath
import time
import threading
import socket

def _datetime_to_epoch(d):
    return mktime(d.timetuple())

class FSHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):

    """A hacked together version of SimpleHTTPRequestHandler"""

    def __init__(self, fs, request, client_address, server):
        self._fs = fs
        SimpleHTTPServer.SimpleHTTPRequestHandler.__init__(self, request, client_address, server)

    def do_GET(self):
        """Serve a GET request."""
        f = None
        try:
            f = self.send_head()
            if f:
                try:
                    self.copyfile(f, self.wfile)
                except socket.error:
                    pass
        finally:
            if f is not None:
                f.close()

    def send_head(self):
        """Common code for GET and HEAD commands.

        This sends the response code and MIME headers.

        Return value is either a file object (which has to be copied
        to the outputfile by the caller unless the command was HEAD,
        and must be closed by the caller under all circumstances), or
        None, in which case the caller has nothing further to do.

        """
        path = self.translate_path(self.path)
        f = None
        if self._fs.isdir(path):
            if not self.path.endswith('/'):
                # redirect browser - doing basically what apache does
                self.send_response(301)
                self.send_header("Location", self.path + "/")
                self.end_headers()
                return None
            for index in ("index.html", "index.htm"):
                index = pathjoin(path, index)
                if self._fs.exists(index):
                    path = index
                    break
            else:
                return self.list_directory(path)
        ctype = self.guess_type(path)
        try:
            info = self._fs.getinfo(path)
            f = self._fs.open(path, 'rb')
        except FSError, e:
            self.send_error(404, str(e))
            return None
        self.send_response(200)
        self.send_header("Content-type", ctype)
        self.send_header("Content-Length", str(info['size']))
        if 'modified_time' in info:
            self.send_header("Last-Modified", self.date_time_string(_datetime_to_epoch(info['modified_time'])))
        self.end_headers()
        return f


    def list_directory(self, path):
        """Helper to produce a directory listing (absent index.html).

        Return value is either a file object, or None (indicating an
        error).  In either case, the headers are sent, making the
        interface the same as for send_head().

        """
        try:
            dir_paths = self._fs.listdir(path, dirs_only=True)
            file_paths = self._fs.listdir(path, files_only=True)
        except FSError:
            self.send_error(404, "No permission to list directory")
            return None
        paths = [p+'/' for p in sorted(dir_paths, key=lambda p:p.lower())] + sorted(file_paths, key=lambda p:p.lower())
        #list.sort(key=lambda a: a.lower())
        f = StringIO()
        displaypath = cgi.escape(urllib.unquote(self.path))
        f.write('<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">')
        f.write("<html>\n<title>Directory listing for %s</title>\n" % displaypath)
        f.write("<body>\n<h2>Directory listing for %s</h2>\n" % displaypath)
        f.write("<hr>\n<ul>\n")

        parent = dirname(path)
        if path != parent:
            f.write('<li><a href="%s">../</a></li>' % urllib.quote(parent.rstrip('/') + '/'))

        for path in paths:
            f.write('<li><a href="%s">%s</a>\n'
                    % (urllib.quote(path), cgi.escape(path)))
        f.write("</ul>\n<hr>\n</body>\n</html>\n")
        length = f.tell()
        f.seek(0)
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.send_header("Content-Length", str(length))
        self.end_headers()
        return f

    def translate_path(self, path):
        # abandon query parameters
        path = path.split('?',1)[0]
        path = path.split('#',1)[0]
        path = posixpath.normpath(urllib.unquote(path))
        return path


def serve_fs(fs, address='', port=8000):

    """Serve an FS instance over http

    :param fs: an FS object
    :param address: IP address to serve on
    :param port: port number

    """

    def Handler(request, client_address, server):
        return FSHTTPRequestHandler(fs, request, client_address, server)

    #class ThreadedTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    #    pass
    httpd = SocketServer.TCPServer((address, port), Handler, bind_and_activate=False)
    #httpd = ThreadedTCPServer((address, port), Handler, bind_and_activate=False)
    httpd.allow_reuse_address = True
    httpd.server_bind()
    httpd.server_activate()

    server_thread = threading.Thread(target=httpd.serve_forever)
    server_thread.start()
    try:
        while True:
            time.sleep(0.1)
    except (KeyboardInterrupt, SystemExit):
        httpd.shutdown()

if __name__ == "__main__":

    from fs.osfs import OSFS
    serve_fs(OSFS('~/'))


#------------------------------------------------------------------------------
#
#  Copyright (c) 2014, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#------------------------------------------------------------------------------
""" Tests for the ShadowGroup class.
"""

import unittest

from traitsui.api import Group
from traitsui.group import ShadowGroup


class TestShadowGroup(unittest.TestCase):

    def test_creation_sets_shadow_first(self):
        group = Group()
        # We end up with a DelegationError if the 'shadow' trait is not set
        # first.  Initialization order is dependent on dictionary order, which
        # we can't control, so we throw in a good number of other traits to
        # increase the chance that some other trait is set first.
        shadow_group = ShadowGroup(
            label='dummy',
            show_border=True,
            show_labels=True,
            show_left=True,
            orientation='horizontal',
            scrollable=True,
            shadow=group,
        )



#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the Boolean editor factory for all traits toolkit backends.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Dict, Str, Any

# CIRCULAR IMPORT FIXME: Importing from the source rather than traits.ui.api
# to avoid circular imports, as this EditorFactory will be part of
# traits.ui.api as well.

from ..view import View

from .text_editor import ToolkitEditorFactory as EditorFactory

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Mapping from user input text to Boolean values
mapping_trait = Dict(Str, Any, {'True': True,
                                'true': True,
                                't': True,
                                'yes': True,
                                'y': True,
                                'False': False,
                                'false': False,
                                'f': False,
                                'no': False,
                                'n': False,
                                })

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for Boolean editors.
    """
    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Dictionary mapping user input to other values.
    # These definitions override definitions in the 'text_editor' version
    mapping = mapping_trait

    #-------------------------------------------------------------------------
    #  Traits view definition:
    #-------------------------------------------------------------------------

    traits_view = View()

    #-------------------------------------------------------------------------
    #  EditorFactory methods
    #-------------------------------------------------------------------------

    def _get_custom_editor_class(self):
        """ Returns the editor class to use for "custom" style views.
        Overridden to return the simple_editor_class (instead of the
        CustomEditor class for the text editor's factory, which this class
        inherits from).

        """
        return self.simple_editor_class


# Define the BooleanEditor class
BooleanEditor = ToolkitEditorFactory

#- EOF -----------------------------------------------------------------------



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/07/2004
#
#------------------------------------------------------------------------------

""" Exports the symbols defined by the traits.ui package.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from .basic_editor_factory import BasicEditorFactory

from .context_value import CV, CVFloat, CVInt, CVStr, CVType, ContextValue

from .editor import Editor

from .editor_factory import EditorFactory

try:
    from .editors.api import ArrayEditor
except ImportError:
    # ArrayEditor depends on numpy, so ignore if numpy is not present.
    pass

from .editors.api import (
    BooleanEditor,
    ButtonEditor,
    CheckListEditor,
    CodeEditor,
    ColorEditor,
    CompoundEditor,
    CustomEditor,
    CSVListEditor,
    DNDEditor,
    StyledDateEditor,
    DateEditor,
    DefaultOverride,
    DirectoryEditor,
    DropEditor,
    EnumEditor,
    FileEditor,
    FontEditor,
    HTMLEditor,
    HistoryEditor,
    ImageEditor,
    ImageEnumEditor,
    InstanceEditor,
    KeyBindingEditor,
    ListEditor,
    ListStrEditor,
    NullEditor,
    PopupEditor,
    ProgressEditor,
    RGBColorEditor,
    RangeEditor,
    ScrubberEditor,
    SearchEditor,
    SetEditor,
    ShellEditor,
    TableEditor,
    TabularEditor,
    TextEditor,
    TimeEditor,
    TitleEditor,
    TreeEditor,
    TupleEditor,
    ValueEditor)

from .group import (Group, HFlow, HGroup, HSplit, Tabbed, VFlow, VFold, VGrid,
                    VGroup, VSplit)

from .handler import Controller, Handler, ModelView, ViewHandler, default_handler

from .help import on_help_call

from .help_template import help_template

from .include import Include

from .item import (Custom, Heading, Item, Label, Readonly, Spring, UCustom,
                   UItem, UReadonly, spring)

from .menu import (
    Action,
    ActionGroup,
    ApplyButton,
    CancelButton,
    CloseAction,
    HelpAction,
    HelpButton,
    LiveButtons,
    Menu,
    MenuBar,
    ModalButtons,
    NoButton,
    NoButtons,
    OKButton,
    OKCancelButtons,
    PyFaceAction,
    RedoAction,
    RevertAction,
    RevertButton,
    Separator,
    StandardMenuBar,
    ToolBar,
    UndoAction,
    UndoButton)

from .message import auto_close_message, error, message

from .table_column import (ExpressionColumn, ListColumn, NumericColumn,
                           ObjectColumn, TableColumn)

from .table_filter import (EvalTableFilter, MenuTableFilter, RuleTableFilter,
                           TableFilter)

from .theme import Theme, default_theme

from .toolkit import toolkit

from .toolkit_traits import ColorTrait, FontTrait, RGBColorTrait

from .tree_node import (ITreeNode, ITreeNodeAdapter, MultiTreeNode,
                        ObjectTreeNode, TreeNode, TreeNodeObject)

from .ui import UI

from .ui_info import UIInfo

from .ui_traits import (ATheme, Border, HasBorder, HasMargin, Image, Margin,
                        StatusItem)

from .undo import (AbstractUndoItem, ListUndoItem, UndoHistory,
                   UndoHistoryUndoItem, UndoItem)

from .view import View

from .view_element import ViewElement, ViewSubElement

from . import view_elements

_constants = toolkit().constants()
WindowColor = _constants.get('WindowColor', 0xFFFFFF)


def raise_to_debug():
    """ When we would otherwise silently swallow an exception, call this instead
    to allow people to set the TRAITS_DEBUG environment variable and get the
    exception.
    """
    import os
    if os.getenv('TRAITS_DEBUG') is not None:
        raise



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" A clickable/draggable widget containing an image. """
from __future__ import absolute_import

# Major package imports.
import wx

# Enthought library imports.
from traits.api import Any, Bool, Event

# Local imports.
from .widget import Widget


class ImageWidget(Widget):
    """ A clickable/draggable widget containing an image. """

    #### 'ImageWidget' interface ##############################################

    # The bitmap.
    bitmap = Any

    # Is the widget selected?
    selected = Bool(False)

    #### Events ####

    # A key was pressed while the tree is in focus.
    key_pressed = Event

    # A node has been activated (ie. double-clicked).
    node_activated = Event

    # A drag operation was started on a node.
    node_begin_drag = Event

    # A (non-leaf) node has been collapsed.
    node_collapsed = Event

    # A (non-leaf) node has been expanded.
    node_expanded = Event

    # A left-click occurred on a node.
    node_left_clicked = Event

    # A right-click occurred on a node.
    node_right_clicked = Event

    #### Private interface ####################################################

    _selected = Any

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__ (self, parent, **traits):
        """ Creates a new widget. """

        # Base class constructors.
        super(ImageWidget, self).__init__(**traits)

        # Add some padding around the image.
        size = (self.bitmap.GetWidth() + 10, self.bitmap.GetHeight() + 10)

        # Create the toolkit-specific control.
        self.control = wx.Window(parent, -1, size=size)
        self.control.__tag__ = 'hack'

        self._mouse_over  = False
        self._button_down = False

        # Set up mouse event handlers:
        wx.EVT_ENTER_WINDOW(self.control, self._on_enter_window)
        wx.EVT_LEAVE_WINDOW(self.control, self._on_leave_window)
        wx.EVT_LEFT_DCLICK(self.control, self._on_left_dclick)
        wx.EVT_LEFT_DOWN(self.control, self._on_left_down)
        wx.EVT_LEFT_UP(self.control, self._on_left_up)
        wx.EVT_PAINT(self.control, self._on_paint)

        # Pens used to draw the 'selection' marker:
        # ZZZ: Make these class instances when moved to the wx toolkit code.
        self._selectedPenDark = wx.Pen(
            wx.SystemSettings_GetColour(wx.SYS_COLOUR_3DSHADOW), 1, wx.SOLID
        )

        self._selectedPenLight = wx.Pen(
            wx.SystemSettings_GetColour(wx.SYS_COLOUR_3DHIGHLIGHT), 1, wx.SOLID
        )

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait event handlers #################################################

    def _bitmap_changed(self, bitmap):
        """ Called when the widget's bitmap is changed. """

        if self.control is not None:
            self.control.Refresh()

        return

    def _selected_changed(self, selected):
        """ Called when the selected state of the widget is changed. """

        if selected:
            for control in self.GetParent().GetChildren():
                if hasattr(control, '__tag__'):
                    if control.Selected():
                      control.Selected(False)
                      break

                self.Refresh()

        return

    #### wx event handlers ####################################################

    def _on_enter_window(self, event):
        """ Called when the mouse enters the widget. """

        if self._selected is not None:
            self._mouse_over = True
            self.Refresh()

        return

    def _on_leave_window(self, event):
        """ Called when the mouse leaves the widget. """

        if self._mouse_over:
            self._mouse_over = False
            self.Refresh()

        return

    def _on_left_dclick(self, event):
        """ Called when the left mouse button is double-clicked. """

        #print 'left dclick'

        event.Skip()

        return

    def _on_left_down ( self, event = None ):
        """ Called when the left mouse button goes down on the widget. """

        #print 'left down'

        if self._selected is not None:
            self.CaptureMouse()
            self._button_down = True
            self.Refresh()

        event.Skip()

        return

    def _on_left_up ( self, event = None ):
        """ Called when the left mouse button goes up on the widget. """

        #print 'left up'

        need_refresh = self._button_down
        if need_refresh:
            self.ReleaseMouse()
            self._button_down = False

        if self._selected is not None:
            wdx, wdy = self.GetClientSizeTuple()
            x        = event.GetX()
            y        = event.GetY()
            if (0 <= x < wdx) and (0 <= y < wdy):
                if self._selected != -1:
                    self.Selected( True )
                elif need_refresh:
                    self.Refresh()

                return

        if need_refresh:
            self.Refresh()

        event.Skip()

        return

    def _on_paint ( self, event = None ):
        """ Called when the widget needs repainting. """

        wdc      = wx.PaintDC( self.control )
        wdx, wdy = self.control.GetClientSizeTuple()
        bitmap   = self.bitmap
        bdx      = bitmap.GetWidth()
        bdy      = bitmap.GetHeight()
        wdc.DrawBitmap( bitmap, (wdx - bdx) / 2, (wdy - bdy) / 2, True )

        pens = [ self._selectedPenLight, self._selectedPenDark ]
        bd   = self._button_down
        if self._mouse_over:
            wdc.SetBrush( wx.TRANSPARENT_BRUSH )
            wdc.SetPen( pens[ bd ] )
            wdc.DrawLine( 0, 0, wdx, 0 )
            wdc.DrawLine( 0, 1, 0, wdy )
            wdc.SetPen( pens[ 1 - bd ] )
            wdc.DrawLine( wdx - 1, 1, wdx - 1, wdy )
            wdc.DrawLine( 1, wdy - 1, wdx - 1, wdy - 1 )

        if self._selected == True:
            wdc.SetBrush( wx.TRANSPARENT_BRUSH )
            wdc.SetPen( pens[ bd ] )
            wdc.DrawLine( 1, 1, wdx - 1, 1 )
            wdc.DrawLine( 1, 1, 1, wdy - 1 )
            wdc.DrawLine( 2, 2, wdx - 2, 2 )
            wdc.DrawLine( 2, 2, 2, wdy - 2 )
            wdc.SetPen( pens[ 1 - bd ] )
            wdc.DrawLine( wdx - 2, 2, wdx - 2, wdy - 1 )
            wdc.DrawLine( 2, wdy - 2, wdx - 2, wdy - 2 )
            wdc.DrawLine( wdx - 3, 3, wdx - 3, wdy - 2 )
            wdc.DrawLine( 3, wdy - 3, wdx - 3, wdy - 3 )

        return

#### EOF ######################################################################



#-------------------------------------------------------------------------------
#
#  A n row x m column scatter-plot template defined as a test for the template
#  package.
#
#  Written by: David C. Morrill
#  (based on the original cp.plot geo_scatter_plot.py file)
#
#  Date: 08/01/2007
#
#  (c) Copy 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" A n row x m column scatter-plot template defined as a test for the template
    package.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import Undefined

from traitsui.api \
    import View, VGroup, Item, Label, Theme, TextEditor

from traitsui.wx.themed_slider_editor \
    import ThemedSliderEditor

from traitsui.wx.themed_text_editor \
    import ThemedTextEditor

from enable.api \
    import ColorTrait

from chaco.api \
    import GridPlotContainer, PlotComponent

from chaco.scatter_markers \
    import marker_trait

from apptools.template.api \
    import MutableTemplate, TRange, TStr, TList, TDerived

from enable_editor \
    import EnableEditor

from scatter_plot \
    import ScatterPlot

#-------------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------------

# Template color trait:
TColor = ColorTrait( template = 'copy' )

#-------------------------------------------------------------------------------
#  'ScatterPlotNM' class:
#-------------------------------------------------------------------------------

class ScatterPlotNM ( MutableTemplate ):

    #-- Template Traits --------------------------------------------------------

    # The title of the plot:
    title = TStr( 'NxM Scatter Plots' )

    # The type of marker to use.  This is a mapped trait using strings as the
    # keys:
    marker = marker_trait( template = 'copy', event = 'update' )

    # The pixel size of the marker (doesn't include the thickness of the
    # outline):
    marker_size = TRange( 1, 5, 1, event = 'update' )

    # The thickness, in pixels, of the outline to draw around the marker.  If
    # this is 0, no outline will be drawn.
    line_width = TRange( 0.0, 5.0, 1.0 )

    # The fill color of the marker:
    color = TColor( 'red', event = 'update' )

    # The color of the outline to draw around the marker
    outline_color = TColor( 'black', event = 'update' )

    # The number of rows of plots:
    rows = TRange( 1, 3, 1, event = 'grid' )

    # The number of columns of plots:
    columns = TRange( 1, 5, 1, event = 'grid' )

    # The contained scatter plots:
    scatter_plots = TList( ScatterPlot )

    #-- Derived Traits ---------------------------------------------------------

    plot = TDerived

    #-- Traits UI Views --------------------------------------------------------

    # The scatter plot view:
    template_view = View(
        VGroup(
            Item( 'title',
                  show_label = False,
                  style      = 'readonly',
                  editor     = ThemedTextEditor(
                                 theme = Theme( '@GBB', alignment = 'center' ) )
            ),
            Item( 'plot',
                  show_label = False,
                  resizable  = True,
                  editor     = EnableEditor(),
                  item_theme = Theme( '@GF5', margins = 0 )
            )
        ),
        resizable = True
    )

    # The scatter plot options view:
    options_view = View(
        VGroup(
            VGroup(
                Label( 'Scatter Plot Options',
                       item_theme = Theme( '@GBB', alignment = 'center' ) ),
                show_labels = False
            ),
            VGroup(
                Item( 'title', editor = TextEditor() ),
                Item( 'marker' ),
                Item( 'marker_size', editor = ThemedSliderEditor() ),
                Item( 'line_width',
                      label  = 'Line Width',
                      editor = ThemedSliderEditor() ),
                Item( 'color',         label = 'Fill Color' ),
                Item( 'outline_color', label = 'Outline Color' ),
                Item( 'rows',    editor = ThemedSliderEditor() ),
                Item( 'columns', editor = ThemedSliderEditor() ),
                group_theme = Theme( '@GF5', margins = ( -5, -1 ) ),
                item_theme  = Theme( '@G0B', margins = 0 )
            )
        )
    )

    #-- ITemplate Interface Implementation -------------------------------------

    def activate_template ( self ):
        """ Converts all contained 'TDerived' objects to real objects using the
            template traits of the object. This method must be overridden in
            subclasses.

            Returns
            -------
            None
        """
        plots = []
        i     = 0
        for r in range( self.rows ):
            row = []
            for c in range( self.columns ):
                plot = self.scatter_plots[i].plot
                if plot is None:
                    plot = PlotComponent()
                row.append( plot )
                i += 1
            plots.append( row )

        self.plot = GridPlotContainer( shape = ( self.rows, self.columns ) )
        self.plot.component_grid = plots

    #-- Default Values ---------------------------------------------------------

    def _scatter_plots_default ( self ):
        """ Returns the default value for the scatter plots list.
        """
        plots = []
        for i in range( self.rows * self.columns ):
            plots.append( ScatterPlot() )

        self._update_plots( plots )

        return plots

    #-- Trait Event Handlers ---------------------------------------------------

    def _update_changed ( self, name, old, new ):
        """ Handles a plot option being changed.
        """
        for sp in self.scatter_plots:
            setattr( sp, name, new )

        self.plot = Undefined

    def _grid_changed ( self ):
        """ Handles the grid size being changed.
        """
        n     = self.rows * self.columns
        plots = self.scatter_plots
        if n < len( plots ):
            self.scatter_plots = plots[:n]
        else:
            for j in range( len( plots ), n ):
                plots.append( ScatterPlot() )

        self._update_plots( plots )

        self.template_mutated = True

    #-- Private Methods --------------------------------------------------------

    def _update_plots ( self, plots ):
        """ Update the data sources for all of the current plots.
        """
        index = None
        i     = 0
        for r in range( self.rows ):
            for c in range( self.columns ):
                sp   = plots[i]
                i   += 1
                desc = sp.value.description
                col  = desc.rfind( '[' )
                if col >= 0:
                    desc = desc[:col]
                sp.value.description = '%s[%d,%d]' % ( desc, r, c )
                sp.value.optional    = True

                if index is None:
                    index = sp.index
                    index.description = 'Shared Plot Index'
                    index.optional    = True
                else:
                     sp.index = index




import math
import re
import string

import OpenGL.GL as gl
import OpenGL.GLU as glu
from svg_constants import *
from svg_parser_utils import *
from vector_math import Matrix
import svg_style
import svg_constants

POINT_RE = re.compile("(-?[0-9]+\.?[0-9]*(?:e-?[0-9]*)?)")
PATH_CMD_RE = re.compile("([A-Za-z]|-?[0-9]+\.?[0-9]*(?:e-?[0-9]*)?)")


class SVGPathBuilder(object):

    def __init__(self, fill_rule='nonzero'):
        self._bezier_coefficients = []
        self.cursor_x = 0
        self.cursor_y = 0
        self.close_index = 0
        self.ctx_path = []
        self.ctx_loop = []
        self.shape = None
        self.fill_rule = fill_rule
        self.n_bezier_points = svg_constants.BEZIER_POINTS
        self.n_circle_points = svg_constants.CIRCLE_POINTS
        self.tolerance = svg_constants.TOLERANCE
        self.fill_rule = fill_rule

    def read_xml_svg_element(self, path, element, config):
        self._bezier_coefficients = []
        self.cursor_x = 0
        self.cursor_y = 0
        self.close_index = 0
        self.ctx_path = []
        self.ctx_loop = []
        self.shape = None
        self.n_bezier_points = config.bezier_points
        self.n_circle_points = config.circle_points
        self.tolerance = config.tolerance
        self.fill_rule = None
        if path.style.fill:
            self.fill_rule = path.style.fill_rule

        e = element
        if e.tag.endswith('path'):
            self.shape = path.shape = 'path'
            self._read_path_commands(e)
        elif e.tag.endswith('rect'):
            self.shape = path.shape = 'rect'
            x = parse_float(e.get('x', '0'))
            y = parse_float(e.get('y', '0'))
            h = parse_float(e.get('height'))
            w = parse_float(e.get('width'))

            rx = parse_float(e.get('rx', '0'))
            ry = parse_float(e.get('ry', str(rx)))
            path.x, path.y, path.w, path.h = x, y, w, h

            if rx == 0 and ry == 0:
                # no rounding, so just draw a simple rectangle
                self.set_cursor_position(x, y)
                self.line_to(x + w, y)
                self.line_to(x + w, y + h)
                self.line_to(x, y + h)
                self.line_to(x, y)
                self.end_path()
            else:
                # rounded rectangle, do some stuff with arcs
                self.set_cursor_position(x, y + ry)
                self.arc_to(rx, ry, 0, 0, 1, x + rx, y)
                self.line_to(x + (w - rx), y)
                self.arc_to(rx, ry, 0, 0, 1, x + w, y + ry)
                self.line_to(x + w, y + (h - ry))
                self.arc_to(rx, ry, 0, 0, 1, x + (w - rx), y+h)
                self.line_to(x + rx, y + h)
                self.arc_to(rx, ry, 0, 0, 1, x, y + (h - ry))
                self.line_to(x, y + ry)
                self.end_path()

        elif e.tag.endswith('polyline') or e.tag.endswith('polygon'):
            if e.tag.endswith('polyline'):
                self.shape = path.shape = 'polyline'
            else:
                self.shape = path.shape = 'polygon'
            path_data = e.get('points')
            path_data = POINT_RE.findall(path_data)

            def next_point():
                return float(path_data.pop(0)), float(path_data.pop(0))
            while path_data:
                self.line_to(*next_point())
            if e.tag.endswith('polygon'):
                self.close_path()
            self.end_path()
        elif e.tag.endswith('line'):
            self.shape = path.shape = 'line'
            x1 = float(e.get('x1'))
            y1 = float(e.get('y1'))
            x2 = float(e.get('x2'))
            y2 = float(e.get('y2'))
            path.x1, path.x1, path.x2, path.y2 = x1, y1, x2, y2
            self.set_cursor_position(x1, y1)
            self.line_to(x2, y2)
            self.end_path()
        elif e.tag.endswith('circle'):
            self.shape = path.shape = 'circle'
            cx = float(e.get('cx', 0))
            cy = float(e.get('cy', 0))
            r = float(e.get('r'))
            path.cx, path.cy, path.r = cx, cy, r
            for i in xrange(config.circle_points):
                theta = 2 * i * math.pi / config.circle_points
                self.line_to(cx + r * math.cos(theta), cy + r * math.sin(theta))
            self.close_path()
            self.end_path()
        elif e.tag.endswith('ellipse'):
            self.shape = path.shape = 'ellipse'
            cx = float(e.get('cx', 0))
            cy = float(e.get('cy', 0))
            rx = float(e.get('rx'))
            ry = float(e.get('ry'))
            path.cx, path.cy, path.rx, path.ry = cx, cy, rx, ry
            for i in xrange(config.circle_points):
                theta = 2 * i * math.pi / config.circle_points
                self.line_to(cx + rx * math.cos(theta), cy + ry * math.sin(theta))
            self.close_path()
            self.end_path()

    def close_path(self):
        self.ctx_loop.append(self.ctx_loop[0][:])
        self.ctx_path.append(self.ctx_loop)
        self.ctx_loop = []

    def set_cursor_position(self, x, y):
        self.cursor_x = x
        self.cursor_y = y
        self.ctx_loop.append([x, y])

    def _read_path_commands(self, e):
        path_data = e.get('d', '')
        path_data = PATH_CMD_RE.findall(path_data)

        def next_point():
            return float(path_data.pop(0)), float(path_data.pop(0))

        opcode = ''
        while path_data:
            prev_opcode = opcode
            if path_data[0] in string.letters:
                opcode = path_data.pop(0)
            else:
                opcode = prev_opcode

            if opcode == 'M':
                self.set_cursor_position(*next_point())
            elif opcode == 'm':
                mx, my = next_point()
                self.set_cursor_position(self.cursor_x + mx, self.cursor_y + my)
            elif opcode == 'Q':  # absolute quadratic curve
                self.quadratic_curve_to(*(next_point() + next_point()))
            elif opcode == 'q':  # relative quadratic curve
                ax, ay = next_point()
                bx, by = next_point()
                self.quadratic_curve_to(
                    ax + self.cursor_x, ay + self.cursor_y,
                    bx + self.cursor_x, by + self.cursor_y)

            elif opcode == 'T':
                # quadratic curve with control point as reflection
                mx = 2 * self.cursor_x - self.last_cx
                my = 2 * self.cursor_y - self.last_cy
                x, y = next_point()
                self.quadratic_curve_to(mx, my, x, y)

            elif opcode == 't':
                # relative quadratic curve with control point as reflection
                mx = 2 * self.cursor_x - self.last_cx
                my = 2 * self.cursor_y - self.last_cy
                x, y = next_point()
                self.quadratic_curve_to(
                    mx + self.cursor_x,
                    my + self.cursor_y,
                    x + self.cursor_x,
                    y + self.cursor_y)

            elif opcode == 'C':
                self.curve_to(*(next_point() + next_point() + next_point()))
            elif opcode == 'c':
                mx, my = self.cursor_x, self.cursor_y
                x1, y1 = next_point()
                x2, y2 = next_point()
                x, y = next_point()

                self.curve_to(mx + x1, my + y1, mx + x2, my + y2, mx + x, my + y)
            elif opcode == 'S':
                self.curve_to(2 * self.cursor_x - self.last_cx, 2 * self.cursor_y - self.last_cy,
                               *(next_point() + next_point()))
            elif opcode == 's':
                mx = self.cursor_x
                my = self.cursor_y
                x1, y1 = 2 * self.cursor_x - self.last_cx, 2 * self.cursor_y - self.last_cy
                x2, y2 = next_point()
                x, y = next_point()

                self.curve_to(x1, y1, mx + x2, my + y2, mx + x, my + y)
            elif opcode == 'A':
                rx, ry = next_point()
                phi = float(path_data.pop(0))
                large_arc = int(path_data.pop(0))
                sweep = int(path_data.pop(0))
                x, y = next_point()
                self.arc_to(rx, ry, phi, large_arc, sweep, x, y)
            elif opcode == 'a':  # relative arc
                rx, ry = next_point()
                phi = float(path_data.pop(0))
                large_arc = int(path_data.pop(0))
                sweep = int(path_data.pop(0))
                x, y = next_point()
                self.arc_to(rx, ry, phi, large_arc, sweep, self.cursor_x + x, self.cursor_y + y)
            elif opcode in 'zZ':
                self.close_path()
            elif opcode == 'L':
                self.line_to(*next_point())
            elif opcode == 'l':
                x, y = next_point()
                self.line_to(self.cursor_x + x, self.cursor_y + y)
            elif opcode == 'H':
                x = float(path_data.pop(0))
                self.line_to(x, self.cursor_y)
            elif opcode == 'h':
                x = float(path_data.pop(0))
                self.line_to(self.cursor_x + x, self.cursor_y)
            elif opcode == 'V':
                y = float(path_data.pop(0))
                self.line_to(self.cursor_x, y)
            elif opcode == 'v':
                y = float(path_data.pop(0))
                self.line_to(self.cursor_x, self.cursor_y + y)
            else:
                self._warn("Unrecognised opcode: " + opcode)
                raise Exception("Unrecognised opcode: " + opcode)
        self.end_path()

    def arc_to(self, rx, ry, phi, large_arc, sweep, x, y):
        # This function is made out of magical fairy dust
        # http://www.w3.org/TR/2003/REC-SVG11-20030114/implnote.html#ArcImplementationNotes
        x1 = self.cursor_x
        y1 = self.cursor_y
        x2 = x
        y2 = y
        cp = math.cos(phi)
        sp = math.sin(phi)
        dx = .5 * (x1 - x2)
        dy = .5 * (y1 - y2)
        x_ = cp * dx + sp * dy
        y_ = -sp * dx + cp * dy
        r2 = (((rx * ry) ** 2 - (rx * y_) ** 2 - (ry * x_) ** 2) /
              ((rx * y_) ** 2 + (ry * x_) ** 2))
        if r2 < 0: r2 = 0
        r = math.sqrt(r2)
        if large_arc == sweep:
            r = -r
        cx_ = r * rx * y_ / ry
        cy_ = -r * ry * x_ / rx
        cx = cp * cx_ - sp * cy_ + .5 * (x1 + x2)
        cy = sp * cx_ + cp * cy_ + .5 * (y1 + y2)

        def angle(u, v):
            a = math.acos((u[0] * v[0] + u[1] * v[1]) / math.sqrt((u[0] ** 2 + u[1] ** 2) * (v[0] ** 2 + v[1] ** 2)))
            sgn = 1 if u[0] * v[1] > u[1] * v[0] else -1
            return sgn * a

        psi = angle((1, 0), ((x_ - cx_) / rx, (y_ - cy_) / ry))
        delta = angle(((x_ - cx_) / rx, (y_ - cy_) / ry),
                      ((-x_ - cx_) / rx, (-y_ - cy_) / ry))
        if sweep and delta < 0:
            delta += math.pi * 2
        if not sweep and delta > 0:
            delta -= math.pi * 2
        n_points = max(int(abs(self.n_circle_points * delta / (2 * math.pi))), 1)

        for i in xrange(n_points + 1):
            theta = psi + i * delta / n_points
            ct = math.cos(theta)
            st = math.sin(theta)
            self.line_to(cp * rx * ct - sp * ry * st + cx,
                         sp * rx * ct + cp * ry * st + cy)

    def quadratic_curve_to(self, x1, y1, x2, y2):
        x0, y0 = self.cursor_x, self.cursor_y
        n_bezier_points = self.n_bezier_points
        for i in xrange(n_bezier_points + 1):
            t = float(i) / n_bezier_points
            q0x = (x1 - x0) * t + x0
            q0y = (y1 - y0) * t + y0

            q1x = (x2 - x1) * t + x1
            q1y = (y2 - y1) * t + y1

            bx = (q1x - q0x) * t + q0x
            by = (q1y - q0y) * t + q0y

            self.ctx_loop.append([bx, by])

        self.last_cx, self.last_cy = x1, y1
        self.cursor_x, self.cursor_y = x2, y2

    def curve_to(self, x1, y1, x2, y2, x, y):
        n_bezier_points = self.n_bezier_points
        if not self._bezier_coefficients:
            for i in xrange(n_bezier_points + 1):
                t = float(i) / n_bezier_points
                t0 = (1 - t) ** 3
                t1 = 3 * t * (1 - t) ** 2
                t2 = 3 * t ** 2 * (1 - t)
                t3 = t ** 3
                self._bezier_coefficients.append([t0, t1, t2, t3])
        self.last_cx = x2
        self.last_cy = y2
        for i, t in enumerate(self._bezier_coefficients):
            px = t[0] * self.cursor_x + t[1] * x1 + t[2] * x2 + t[3] * x
            py = t[0] * self.cursor_y + t[1] * y1 + t[2] * y2 + t[3] * y
            self.ctx_loop.append([px, py])

        self.cursor_x, self.cursor_y = px, py

    def line_to(self, x, y):
        self.set_cursor_position(x, y)

    def end_path(self):
        self.ctx_path.append(self.ctx_loop)
        if self.ctx_path:
            path = []
            for orig_loop in self.ctx_path:
                if not orig_loop:
                    continue
                loop = [orig_loop[0]]
                for pt in orig_loop:
                    if (pt[0] - loop[-1][0]) ** 2 + (pt[1] - loop[-1][1])**2 > self.tolerance:
                        loop.append(pt)
                path.append(loop)

            self.path = path
            self.polygon = self._triangulate(path, self.fill_rule) if self.fill_rule else None
        self.ctx_path = []

        return self.path, self.polygon

    def _triangulate(self, looplist, fill_rule):
        if self.shape in ['line']:
            return None
        t_list = []
        self.ctx_curr_shape = []
        spare_verts = []
        tess = glu.gluNewTess()
        glu.gluTessNormal(tess, 0, 0, 1)
        glu.gluTessProperty(tess, glu.GLU_TESS_WINDING_RULE, glu.GLU_TESS_WINDING_NONZERO)

        def set_tess_callback(which):
            def set_call(func):
                glu.gluTessCallback(tess, which, func)
            return set_call

        @set_tess_callback(glu.GLU_TESS_VERTEX)
        def vertex_callback(vertex):
            self.ctx_curr_shape.append(list(vertex[0:2]))

        @set_tess_callback(glu.GLU_TESS_BEGIN)
        def begin_callback(which):
            self.ctx_tess_style = which

        @set_tess_callback(glu.GLU_TESS_END)
        def end_callback():
            if self.ctx_tess_style == gl.GL_TRIANGLE_FAN:
                c = self.ctx_curr_shape.pop(0)
                p1 = self.ctx_curr_shape.pop(0)
                while self.ctx_curr_shape:
                    p2 = self.ctx_curr_shape.pop(0)
                    t_list.extend([c, p1, p2])
                    p1 = p2
            elif self.ctx_tess_style == gl.GL_TRIANGLE_STRIP:
                p1 = self.ctx_curr_shape.pop(0)
                p2 = self.ctx_curr_shape.pop(0)
                while self.ctx_curr_shape:
                    p3 = self.ctx_curr_shape.pop(0)
                    t_list.extend([p1, p2, p3])
                    p1 = p2
                    p2 = p3
            elif self.ctx_tess_style == gl.GL_TRIANGLES:
                t_list.extend(self.ctx_curr_shape)
            else:
                self._warn("Unrecognised tesselation style: %d" % (self.ctx_tess_style,))
            self.ctx_tess_style = None
            self.ctx_curr_shape = []

        @set_tess_callback(glu.GLU_TESS_ERROR)
        def error_callback(code):
            ptr = glu.gluErrorString(code)
            err = ''
            idx = 0
            while ptr[idx]:
                err += chr(ptr[idx])
                idx += 1
            self._warn("GLU Tesselation Error: " + err)

        @set_tess_callback(glu.GLU_TESS_COMBINE)
        def combine_callback(coords, vertex_data, weights):
            x, y, z = coords[0:3]
            dataOut = (x,y,z)
            spare_verts.append((x,y,z))
            return dataOut

        data_lists = []
        for vlist in looplist:
            d_list = []
            for x, y in vlist:
                v_data = (x, y, 0)
                d_list.append(v_data)
            data_lists.append(d_list)

        if fill_rule == 'nonzero':
            glu.gluTessProperty(tess, glu.GLU_TESS_WINDING_RULE, glu.GLU_TESS_WINDING_NONZERO)
        elif fill_rule == 'evenodd':
            glu.gluTessProperty(tess, glu.GLU_TESS_WINDING_RULE, glu.GLU_TESS_WINDING_ODD)

        glu.gluTessBeginPolygon(tess, None)
        for d_list in data_lists:
            glu.gluTessBeginContour(tess)
            for v_data in d_list:
                glu.gluTessVertex(tess, v_data, v_data)
            glu.gluTessEndContour(tess)
        glu.gluTessEndPolygon(tess)
        return t_list

    def _warn(self, message):
        print "Warning: SVG Parser - %s" % (message,)







import urlparse
import mimetypes

from fs.errors import FSError
from fs.path import basename, pathsplit

from datetime import datetime

try:
    from mako.template import Template
except ImportError:
    print "Requires mako templates http://www.makotemplates.org/"
    raise


class Request(object):
    """Very simple request object"""
    def __init__(self, environ, start_response):
        self.environ = environ
        self.start_response = start_response
        self.path = environ.get('PATH_INFO')


class WSGIServer(object):
    """Light-weight WSGI server that exposes an FS"""

    def __init__(self, serve_fs, indexes=True, dir_template=None, chunk_size=16*1024*1024):

        if dir_template is None:
            from dirtemplate import template as dir_template

        self.serve_fs = serve_fs
        self.indexes = indexes
        self.chunk_size = chunk_size

        self.dir_template = Template(dir_template)

    def __call__(self, environ, start_response):

        request = Request(environ, start_response)

        if not self.serve_fs.exists(request.path):
            return self.serve_404(request)

        if self.serve_fs.isdir(request.path):
            if not self.indexes:
                return self.serve_404(request)
            return self.serve_dir(request)
        else:
            return self.serve_file(request)


    def serve_file(self, request):
        """Serve a file, guessing a mime-type"""
        path = request.path
        serving_file = None
        try:
            serving_file = self.serve_fs.open(path, 'rb')
        except Exception, e:
            if serving_file is not None:
                serving_file.close()
            return self.serve_500(request, str(e))

        mime_type = mimetypes.guess_type(basename(path))[0] or b'text/plain'
        file_size = self.serve_fs.getsize(path)
        headers = [(b'Content-Type', bytes(mime_type)),
                   (b'Content-Length', bytes(file_size))]

        def gen_file():
            chunk_size = self.chunk_size
            read = serving_file.read
            try:
                while 1:
                    data = read(chunk_size)
                    if not data:
                        break
                    yield data
            finally:
                serving_file.close()

        request.start_response(b'200 OK',
                               headers)
        return gen_file()

    def serve_dir(self, request):
        """Serve an index page"""
        fs = self.serve_fs
        isdir = fs.isdir
        path = request.path
        dirinfo = fs.listdirinfo(path, full=True, absolute=True)
        entries = []

        for p, info in dirinfo:
            entry = {}
            entry['path'] = p
            entry['name'] = basename(p)
            entry['size'] = info.get('size', 'unknown')
            entry['created_time'] = info.get('created_time')
            if isdir(p):
                entry['type'] = 'dir'
            else:
                entry['type'] = 'file'

            entries.append(entry)

        # Put dirs first, and sort by reverse created time order
        no_time = datetime(1970, 1, 1, 1, 0)
        entries.sort(key=lambda k:(k['type'] == 'dir', k.get('created_time') or no_time), reverse=True)

        # Turn datetime to text and tweak names
        for entry in entries:
            t = entry.get('created_time')
            if t and hasattr(t, 'ctime'):
                entry['created_time'] = t.ctime()
            if entry['type'] == 'dir':
                entry['name'] += '/'

        # Add an up dir link for non-root
        if path not in ('', '/'):
            entries.insert(0, dict(name='../', path='../', type="dir", size='', created_time='..'))

        # Render the mako template
        html = self.dir_template.render(**dict(fs=self.serve_fs,
                                               path=path,
                                               dirlist=entries)).encode('utf-8')
        request.start_response(b'200 OK', [(b'Content-Type', b'text/html'),
                                           (b'Content-Length', b'%i' % len(html))])

        return [html]


    def serve_404(self, request, msg='Not found'):
        """Serves a Not found page"""
        request.start_response(b'404 NOT FOUND', [(b'Content-Type', b'text/html')])
        return [msg]

    def serve_500(self, request, msg='Unable to complete request'):
        """Serves an internal server error page"""
        request.start_response(b'500 INTERNAL SERVER ERROR', [(b'Content-Type', b'text/html')])
        return [msg]


def serve_fs(fs, indexes=True):
    """Serves an FS object via WSGI"""
    application = WSGIServer(fs, indexes)
    return application




#!/usr/bin/env python
"""
fs.memoryfs
===========

A Filesystem that exists in memory only. Which makes them extremely fast, but non-permanent.

If you open a file from a `memoryfs` you will get back a StringIO object from the standard library.


"""

import datetime
import stat
from fs.path import iteratepath, pathsplit, normpath
from fs.base import *
from fs.errors import *
from fs import _thread_synchronize_default
from fs.filelike import StringIO
from fs import iotools
from os import SEEK_END
import threading

import six
from six import b


def _check_mode(mode, mode_chars):
    for c in mode_chars:
        if c not in mode:
            return False
    return True


class MemoryFile(object):

    def seek_and_lock(f):
        def deco(self, *args, **kwargs):
            try:
                self._lock.acquire()
                self.mem_file.seek(self.pos)
                ret = f(self, *args, **kwargs)
                self.pos = self.mem_file.tell()
                return ret
            finally:
                self._lock.release()
        return deco

    def __init__(self, path, memory_fs, mem_file, mode, lock):
        self.closed = False
        self.path = path
        self.memory_fs = memory_fs
        self.mem_file = mem_file
        self.mode = mode
        self._lock = lock

        self.pos = 0

        if _check_mode(mode, 'a'):
            lock.acquire()
            try:
                self.mem_file.seek(0, SEEK_END)
                self.pos = self.mem_file.tell()
            finally:
                lock.release()

        elif _check_mode(mode, 'w'):
            lock.acquire()
            try:
                self.mem_file.seek(0)
                self.mem_file.truncate()
            finally:
                lock.release()

        assert self.mem_file is not None, "self.mem_file should have a value"

    def __str__(self):
        return "<MemoryFile in %s %s>" % (self.memory_fs, self.path)

    def __repr__(self):
        return u"<MemoryFile in %s %s>" % (self.memory_fs, self.path)

    def __unicode__(self):
        return u"<MemoryFile in %s %s>" % (self.memory_fs, self.path)

    def __del__(self):
        if not self.closed:
            self.close()

    def flush(self):
        pass

    def __iter__(self):
        if 'r' not in self.mode and '+' not in self.mode:
            raise IOError("File not open for reading")
        self.mem_file.seek(self.pos)
        for line in self.mem_file:
            yield line

    @seek_and_lock
    def next(self):
        if 'r' not in self.mode and '+' not in self.mode:
            raise IOError("File not open for reading")
        return self.mem_file.next()

    @seek_and_lock
    def readline(self, *args, **kwargs):
        if 'r' not in self.mode and '+' not in self.mode:
            raise IOError("File not open for reading")
        return self.mem_file.readline(*args, **kwargs)

    def close(self):
        do_close = False
        self._lock.acquire()
        try:
            do_close = not self.closed and self.mem_file is not None
            if do_close:
                self.closed = True
        finally:
            self._lock.release()
        if do_close:
            self.memory_fs._on_close_memory_file(self, self.path)

    @seek_and_lock
    def read(self, size=None):
        if 'r' not in self.mode and '+' not in self.mode:
            raise IOError("File not open for reading")
        if size is None:
            size = -1
        return self.mem_file.read(size)

    @seek_and_lock
    def seek(self, *args, **kwargs):
        return self.mem_file.seek(*args, **kwargs)

    @seek_and_lock
    def tell(self):
        return self.pos

    @seek_and_lock
    def truncate(self, *args, **kwargs):
        if 'r' in self.mode and '+' not in self.mode:
            raise IOError("File not open for writing")
        return self.mem_file.truncate(*args, **kwargs)

    #@seek_and_lock
    def write(self, data):
        if 'r' in self.mode and '+' not in self.mode:
            raise IOError("File not open for writing")
        self.memory_fs._on_modify_memory_file(self.path)
        self._lock.acquire()
        try:
            self.mem_file.seek(self.pos)
            self.mem_file.write(data)
            self.pos = self.mem_file.tell()
        finally:
            self._lock.release()

    @seek_and_lock
    def writelines(self, *args, **kwargs):
        return self.mem_file.writelines(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return False


class DirEntry(object):

    def sync(f):
        def deco(self, *args, **kwargs):
            if self.lock is not None:
                try:
                    self.lock.acquire()
                    return f(self, *args, **kwargs)
                finally:
                    self.lock.release()
            else:
                return f(self, *args, **kwargs)
        return deco

    def __init__(self, type, name, contents=None):

        assert type in ("dir", "file"), "Type must be dir or file!"

        self.type = type
        self.name = name

        if contents is None and type == "dir":
            contents = {}

        self.open_files = []
        self.contents = contents
        self.mem_file = None
        self.created_time = datetime.datetime.now()
        self.modified_time = self.created_time
        self.accessed_time = self.created_time

        self.xattrs = {}

        self.lock = None
        if self.type == 'file':
            self.mem_file = StringIO()
            self.lock = threading.RLock()

    def get_value(self):
        self.lock.acquire()
        try:
            return self.mem_file.getvalue()
        finally:
            self.lock.release()
    data = property(get_value)

    def desc_contents(self):
        if self.isfile():
            return "<file %s>" % self.name
        elif self.isdir():
            return "<dir %s>" % "".join("%s: %s" % (k, v.desc_contents()) for k, v in self.contents.iteritems())

    def isdir(self):
        return self.type == "dir"

    def isfile(self):
        return self.type == "file"

    def __str__(self):
        return "%s: %s" % (self.name, self.desc_contents())

    @sync
    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop('lock')
        if self.mem_file is not None:
            state['mem_file'] = self.data
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        if self.type == 'file':
            self.lock = threading.RLock()
        else:
            self.lock = None
        if self.mem_file is not None:
            data = self.mem_file
            self.mem_file = StringIO()
            self.mem_file.write(data)


class MemoryFS(FS):
    """An in-memory filesystem.

    """

    _meta = {'thread_safe': True,
             'network': False,
             'virtual': False,
             'read_only': False,
             'unicode_paths': True,
             'case_insensitive_paths': False,
             'atomic.move': False,
             'atomic.copy': False,
             'atomic.makedir': True,
             'atomic.rename': True,
             'atomic.setcontents': False}

    def _make_dir_entry(self, *args, **kwargs):
        return self.dir_entry_factory(*args, **kwargs)

    def __init__(self, file_factory=None):
        super(MemoryFS, self).__init__(thread_synchronize=_thread_synchronize_default)

        self.dir_entry_factory = DirEntry
        self.file_factory = file_factory or MemoryFile
        if not callable(self.file_factory):
            raise ValueError("file_factory should be callable")

        self.root = self._make_dir_entry('dir', 'root')

    def __str__(self):
        return "<MemoryFS>"

    def __repr__(self):
        return "MemoryFS()"

    def __unicode__(self):
        return "<MemoryFS>"

    @synchronize
    def _get_dir_entry(self, dirpath):
        dirpath = normpath(dirpath)
        current_dir = self.root
        for path_component in iteratepath(dirpath):
            if current_dir.contents is None:
                return None
            dir_entry = current_dir.contents.get(path_component, None)
            if dir_entry is None:
                return None
            current_dir = dir_entry
        return current_dir

    @synchronize
    def _dir_entry(self, path):
        dir_entry = self._get_dir_entry(path)
        if dir_entry is None:
            raise ResourceNotFoundError(path)
        return dir_entry

    @synchronize
    def desc(self, path):
        if self.isdir(path):
            return "Memory dir"
        elif self.isfile(path):
            return "Memory file object"
        else:
            return "No description available"

    @synchronize
    def isdir(self, path):
        path = normpath(path)
        if path in ('', '/'):
            return True
        dir_item = self._get_dir_entry(path)
        if dir_item is None:
            return False
        return dir_item.isdir()

    @synchronize
    def isfile(self, path):
        path = normpath(path)
        if path in ('', '/'):
            return False
        dir_item = self._get_dir_entry(path)
        if dir_item is None:
            return False
        return dir_item.isfile()

    @synchronize
    def exists(self, path):
        path = normpath(path)
        if path in ('', '/'):
            return True
        return self._get_dir_entry(path) is not None

    @synchronize
    def makedir(self, dirname, recursive=False, allow_recreate=False):
        if not dirname and not allow_recreate:
            raise PathError(dirname)
        fullpath = normpath(dirname)
        if fullpath in ('', '/'):
            if allow_recreate:
                return
            raise DestinationExistsError(dirname)
        dirpath, dirname = pathsplit(dirname.rstrip('/'))

        if recursive:
            parent_dir = self._get_dir_entry(dirpath)
            if parent_dir is not None:
                if parent_dir.isfile():
                    raise ResourceInvalidError(dirname, msg="Can not create a directory, because path references a file: %(path)s")
                else:
                    if not allow_recreate:
                        if dirname in parent_dir.contents:
                            raise DestinationExistsError(dirname, msg="Can not create a directory that already exists (try allow_recreate=True): %(path)s")

            current_dir = self.root
            for path_component in iteratepath(dirpath)[:-1]:
                dir_item = current_dir.contents.get(path_component, None)
                if dir_item is None:
                    break
                if not dir_item.isdir():
                    raise ResourceInvalidError(dirname, msg="Can not create a directory, because path references a file: %(path)s")
                current_dir = dir_item

            current_dir = self.root
            for path_component in iteratepath(dirpath):
                dir_item = current_dir.contents.get(path_component, None)
                if dir_item is None:
                    new_dir = self._make_dir_entry("dir", path_component)
                    current_dir.contents[path_component] = new_dir
                    current_dir = new_dir
                else:
                    current_dir = dir_item

            parent_dir = current_dir

        else:
            parent_dir = self._get_dir_entry(dirpath)
            if parent_dir is None:
                raise ParentDirectoryMissingError(dirname, msg="Could not make dir, as parent dir does not exist: %(path)s")

        dir_item = parent_dir.contents.get(dirname, None)
        if dir_item is not None:
            if dir_item.isdir():
                if not allow_recreate:
                    raise DestinationExistsError(dirname)
            else:
                raise ResourceInvalidError(dirname, msg="Can not create a directory, because path references a file: %(path)s")

        if dir_item is None:
            parent_dir.contents[dirname] = self._make_dir_entry("dir", dirname)


    #@synchronize
    #def _orphan_files(self, file_dir_entry):
    #    for f in file_dir_entry.open_files[:]:
    #        f.close()


    @synchronize
    @iotools.filelike_to_stream
    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        path = normpath(path)
        filepath, filename = pathsplit(path)
        parent_dir_entry = self._get_dir_entry(filepath)

        if parent_dir_entry is None or not parent_dir_entry.isdir():
            raise ResourceNotFoundError(path)

        if 'r' in mode or 'a' in mode:
            if filename not in parent_dir_entry.contents:
                raise ResourceNotFoundError(path)

            file_dir_entry = parent_dir_entry.contents[filename]
            if file_dir_entry.isdir():
                raise ResourceInvalidError(path)

            file_dir_entry.accessed_time = datetime.datetime.now()

            mem_file = self.file_factory(path, self, file_dir_entry.mem_file, mode, file_dir_entry.lock)
            file_dir_entry.open_files.append(mem_file)
            return mem_file

        elif 'w' in mode:
            if filename not in parent_dir_entry.contents:
                file_dir_entry = self._make_dir_entry("file", filename)
                parent_dir_entry.contents[filename] = file_dir_entry
            else:
                file_dir_entry = parent_dir_entry.contents[filename]

            file_dir_entry.accessed_time = datetime.datetime.now()

            mem_file = self.file_factory(path, self, file_dir_entry.mem_file, mode, file_dir_entry.lock)
            file_dir_entry.open_files.append(mem_file)
            return mem_file

        if parent_dir_entry is None:
            raise ResourceNotFoundError(path)

    @synchronize
    def remove(self, path):
        dir_entry = self._get_dir_entry(path)

        if dir_entry is None:
            raise ResourceNotFoundError(path)

        if dir_entry.isdir():
            raise ResourceInvalidError(path, msg="That's a directory, not a file: %(path)s")

        pathname, dirname = pathsplit(path)
        parent_dir = self._get_dir_entry(pathname)
        del parent_dir.contents[dirname]

    @synchronize
    def removedir(self, path, recursive=False, force=False):
        path = normpath(path)
        if path in ('', '/'):
            raise RemoveRootError(path)
        dir_entry = self._get_dir_entry(path)

        if dir_entry is None:
            raise ResourceNotFoundError(path)
        if not dir_entry.isdir():
            raise ResourceInvalidError(path, msg="Can't remove resource, its not a directory: %(path)s" )

        if dir_entry.contents and not force:
            raise DirectoryNotEmptyError(path)

        if recursive:
            rpathname = path
            while rpathname:
                rpathname, dirname = pathsplit(rpathname)
                parent_dir = self._get_dir_entry(rpathname)
                if not dirname:
                    raise RemoveRootError(path)
                del parent_dir.contents[dirname]
                # stop recursing if the directory has other contents
                if parent_dir.contents:
                    break
        else:
            pathname, dirname = pathsplit(path)
            parent_dir = self._get_dir_entry(pathname)
            if not dirname:
                raise RemoveRootError(path)
            del parent_dir.contents[dirname]

    @synchronize
    def rename(self, src, dst):
        src = normpath(src)
        dst = normpath(dst)
        src_dir, src_name = pathsplit(src)
        src_entry = self._get_dir_entry(src)
        if src_entry is None:
            raise ResourceNotFoundError(src)
        open_files = src_entry.open_files[:]
        for f in open_files:
            f.flush()
            f.path = dst

        dst_dir,dst_name = pathsplit(dst)
        dst_entry = self._get_dir_entry(dst)
        if dst_entry is not None:
            raise DestinationExistsError(dst)

        src_dir_entry = self._get_dir_entry(src_dir)
        src_xattrs = src_dir_entry.xattrs.copy()
        dst_dir_entry = self._get_dir_entry(dst_dir)
        if dst_dir_entry is None:
            raise ParentDirectoryMissingError(dst)
        dst_dir_entry.contents[dst_name] = src_dir_entry.contents[src_name]
        dst_dir_entry.contents[dst_name].name = dst_name
        dst_dir_entry.xattrs.update(src_xattrs)
        del src_dir_entry.contents[src_name]

    @synchronize
    def settimes(self, path, accessed_time=None, modified_time=None):
        now = datetime.datetime.now()
        if accessed_time is None:
            accessed_time = now
        if modified_time is None:
            modified_time = now

        dir_entry = self._get_dir_entry(path)
        if dir_entry is not None:
            dir_entry.accessed_time = accessed_time
            dir_entry.modified_time = modified_time
            return True
        return False

    @synchronize
    def _on_close_memory_file(self, open_file, path):
        dir_entry = self._get_dir_entry(path)
        if dir_entry is not None and open_file in dir_entry.open_files:
            dir_entry.open_files.remove(open_file)


    @synchronize
    def _on_modify_memory_file(self, path):
        dir_entry = self._get_dir_entry(path)
        if dir_entry is not None:
            dir_entry.modified_time = datetime.datetime.now()

    @synchronize
    def listdir(self, path="/", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        dir_entry = self._get_dir_entry(path)
        if dir_entry is None:
            raise ResourceNotFoundError(path)
        if dir_entry.isfile():
            raise ResourceInvalidError(path, msg="not a directory: %(path)s")
        paths = dir_entry.contents.keys()
        for (i,p) in enumerate(paths):
            if not isinstance(p,unicode):
                paths[i] = unicode(p)
        return self._listdir_helper(path, paths, wildcard, full, absolute, dirs_only, files_only)

    @synchronize
    def getinfo(self, path):
        dir_entry = self._get_dir_entry(path)

        if dir_entry is None:
            raise ResourceNotFoundError(path)

        info = {}
        info['created_time'] = dir_entry.created_time
        info['modified_time'] = dir_entry.modified_time
        info['accessed_time'] = dir_entry.accessed_time

        if dir_entry.isdir():
            info['st_mode'] = 0755 | stat.S_IFDIR
        else:
            info['size'] = len(dir_entry.data or b(''))
            info['st_mode'] = 0666 | stat.S_IFREG

        return info

    @synchronize
    def copydir(self, src, dst, overwrite=False, ignore_errors=False, chunk_size=1024*64):
        src_dir_entry = self._get_dir_entry(src)
        if src_dir_entry is None:
            raise ResourceNotFoundError(src)
        src_xattrs = src_dir_entry.xattrs.copy()
        super(MemoryFS, self).copydir(src, dst, overwrite, ignore_errors=ignore_errors, chunk_size=chunk_size)
        dst_dir_entry = self._get_dir_entry(dst)
        if dst_dir_entry is not None:
            dst_dir_entry.xattrs.update(src_xattrs)

    @synchronize
    def movedir(self, src, dst, overwrite=False, ignore_errors=False, chunk_size=1024*64):
        src_dir_entry = self._get_dir_entry(src)
        if src_dir_entry is None:
            raise ResourceNotFoundError(src)
        src_xattrs = src_dir_entry.xattrs.copy()
        super(MemoryFS, self).movedir(src, dst, overwrite, ignore_errors=ignore_errors, chunk_size=chunk_size)
        dst_dir_entry = self._get_dir_entry(dst)
        if dst_dir_entry is not None:
            dst_dir_entry.xattrs.update(src_xattrs)

    @synchronize
    def copy(self, src, dst, overwrite=False, chunk_size=1024*64):
        src_dir_entry = self._get_dir_entry(src)
        if src_dir_entry is None:
            raise ResourceNotFoundError(src)
        src_xattrs = src_dir_entry.xattrs.copy()
        super(MemoryFS, self).copy(src, dst, overwrite, chunk_size)
        dst_dir_entry = self._get_dir_entry(dst)
        if dst_dir_entry is not None:
            dst_dir_entry.xattrs.update(src_xattrs)

    @synchronize
    def move(self, src, dst, overwrite=False, chunk_size=1024*64):
        src_dir_entry = self._get_dir_entry(src)
        if src_dir_entry is None:
            raise ResourceNotFoundError(src)
        src_xattrs = src_dir_entry.xattrs.copy()
        super(MemoryFS, self).move(src, dst, overwrite, chunk_size)
        dst_dir_entry = self._get_dir_entry(dst)
        if dst_dir_entry is not None:
            dst_dir_entry.xattrs.update(src_xattrs)

    @synchronize
    def getcontents(self, path, mode="rb", encoding=None, errors=None, newline=None):
        dir_entry = self._get_dir_entry(path)
        if dir_entry is None:
            raise ResourceNotFoundError(path)
        if not dir_entry.isfile():
            raise ResourceInvalidError(path, msg="not a file: %(path)s")
        data = dir_entry.data or b('')
        if 'b' not in mode:
            return iotools.decode_binary(data, encoding=encoding, errors=errors, newline=newline)
        return data

    @synchronize
    def setcontents(self, path, data=b'', encoding=None, errors=None, chunk_size=1024*64):
        if isinstance(data, six.binary_type):
            if not self.exists(path):
                self.open(path, 'wb').close()
            dir_entry = self._get_dir_entry(path)
            if not dir_entry.isfile():
                raise ResourceInvalidError('Not a directory %(path)s', path)
            new_mem_file = StringIO()
            new_mem_file.write(data)
            dir_entry.mem_file = new_mem_file
            return len(data)

        return super(MemoryFS, self).setcontents(path, data=data, encoding=encoding, errors=errors, chunk_size=chunk_size)

        # if isinstance(data, six.text_type):
        #     return super(MemoryFS, self).setcontents(path, data, encoding=encoding, errors=errors, chunk_size=chunk_size)
        # if not self.exists(path):
        #     self.open(path, 'wb').close()

        # dir_entry = self._get_dir_entry(path)
        # if not dir_entry.isfile():
        #     raise ResourceInvalidError('Not a directory %(path)s', path)
        # new_mem_file = StringIO()
        # new_mem_file.write(data)
        # dir_entry.mem_file = new_mem_file

    @synchronize
    def setxattr(self, path, key, value):
        dir_entry = self._dir_entry(path)
        key = unicode(key)
        dir_entry.xattrs[key] = value

    @synchronize
    def getxattr(self, path, key, default=None):
        key = unicode(key)
        dir_entry = self._dir_entry(path)
        return dir_entry.xattrs.get(key, default)

    @synchronize
    def delxattr(self, path, key):
        dir_entry = self._dir_entry(path)
        try:
            del dir_entry.xattrs[key]
        except KeyError:
            pass

    @synchronize
    def listxattrs(self, path):
        dir_entry = self._dir_entry(path)
        return dir_entry.xattrs.keys()



"""

fs.contrib.bigfs.subrangefile
=============================

A file-like object that allows wrapping of part of a binary file for reading.

This avoids needless copies of data for large binary files if StringIO would
be used.

Written by Koen van de Sande
http://www.tibed.net

Contributed under the terms of the BSD License:
http://www.opensource.org/licenses/bsd-license.php
"""


class SubrangeFile:
    """File-like class with read-only, binary mode restricting access to a subrange of the whole file"""
    def __init__(self, f, startOffset, fileSize):
        if not hasattr(f, 'read'):
            self.f = open(f, "rb")
            self.name = f
        else:
            self.f = f
            self.name = str(f)
        self.startOffset = startOffset
        self.fileSize = fileSize
        self.seek(0)
        
    def __str__(self):
        return "<SubrangeFile: %s@%d size=%d>" % (self.name, self.startOffset, self.fileSize)

    def __unicode__(self):
        return unicode(self.__str__())

    def size(self):
        return self.fileSize

    def seek(self, offset, whence=0):
        if whence == 0:
            offset = self.startOffset + offset
        elif whence == 1:
            offset = self.startOffset + self.tell() + offset
        elif whence == 2:
            if offset > 0:
                offset = 0
            offset = self.startOffset + self.fileSize + offset
        self.f.seek(offset)
        
    def tell(self):
        return self.f.tell() - self.startOffset

    def __maxSize(self,size=None):
        iSize = self.fileSize
        if not size is None:
            if size < iSize:
                iSize = size
        if self.tell() + iSize > self.fileSize:
            iSize = self.fileSize - self.tell()
        return iSize
            
    def readline(self,size=None):
        toRead = self.__maxSize(size)
        return self.f.readline(toRead)

    def read(self,size=None):
        toRead = self.__maxSize(size)
        return self.f.read(toRead)

    def readlines(self,size=None):
        toRead = self.__maxSize(size)
        temp = self.f.readlines(toRead)
        # now cut off more than we should read...
        result = []
        counter = 0
        for line in temp:
            if counter + len(line) > toRead:
                if toRead == counter:
                    break
                result.append(line[0:(toRead-counter)])
                break
            else:
                result.append(line)
                counter += len(line)
        return result





#-------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   09/22/2005
#
#-------------------------------------------------------------------------

""" Converts a wx.KeyEvent to a standardized "name".
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# Mapping from wxPython special key names to Enable key names
key_map = {
    wx.WXK_BACK: 'Backspace',
    wx.WXK_TAB: 'Tab',
    wx.WXK_RETURN: 'Enter',
    wx.WXK_ESCAPE: 'Esc',
    wx.WXK_DELETE: 'Delete',
    wx.WXK_START: 'Start',
    wx.WXK_LBUTTON: 'Left Button',
    wx.WXK_RBUTTON: 'Right Button',
    wx.WXK_CANCEL: 'Cancel',
    wx.WXK_MBUTTON: 'Middle Button',
    wx.WXK_CLEAR: 'Clear',
    wx.WXK_SHIFT: 'Shift',
    wx.WXK_CONTROL: 'Control',
    wx.WXK_MENU: 'Menu',
    wx.WXK_PAUSE: 'Pause',
    wx.WXK_CAPITAL: 'Capital',
    wx.WXK_PRIOR: 'Page Up',
    wx.WXK_NEXT: 'Page Down',
    wx.WXK_END: 'End',
    wx.WXK_HOME: 'Home',
    wx.WXK_LEFT: 'Left',
    wx.WXK_UP: 'Up',
    wx.WXK_RIGHT: 'Right',
    wx.WXK_DOWN: 'Down',
    wx.WXK_SELECT: 'Select',
    wx.WXK_PRINT: 'Print',
    wx.WXK_EXECUTE: 'Execute',
    wx.WXK_SNAPSHOT: 'Snapshot',
    wx.WXK_INSERT: 'Insert',
    wx.WXK_HELP: 'Help',
    wx.WXK_NUMPAD0: 'Numpad 0',
    wx.WXK_NUMPAD1: 'Numpad 1',
    wx.WXK_NUMPAD2: 'Numpad 2',
    wx.WXK_NUMPAD3: 'Numpad 3',
    wx.WXK_NUMPAD4: 'Numpad 4',
    wx.WXK_NUMPAD5: 'Numpad 5',
    wx.WXK_NUMPAD6: 'Numpad 6',
    wx.WXK_NUMPAD7: 'Numpad 7',
    wx.WXK_NUMPAD8: 'Numpad 8',
    wx.WXK_NUMPAD9: 'Numpad 9',
    wx.WXK_MULTIPLY: 'Multiply',
    wx.WXK_ADD: 'Add',
    wx.WXK_SEPARATOR: 'Separator',
    wx.WXK_SUBTRACT: 'Subtract',
    wx.WXK_DECIMAL: 'Decimal',
    wx.WXK_DIVIDE: 'Divide',
    wx.WXK_F1: 'F1',
    wx.WXK_F2: 'F2',
    wx.WXK_F3: 'F3',
    wx.WXK_F4: 'F4',
    wx.WXK_F5: 'F5',
    wx.WXK_F6: 'F6',
    wx.WXK_F7: 'F7',
    wx.WXK_F8: 'F8',
    wx.WXK_F9: 'F9',
    wx.WXK_F10: 'F10',
    wx.WXK_F11: 'F11',
    wx.WXK_F12: 'F12',
    wx.WXK_F13: 'F13',
    wx.WXK_F14: 'F14',
    wx.WXK_F15: 'F15',
    wx.WXK_F16: 'F16',
    wx.WXK_F17: 'F17',
    wx.WXK_F18: 'F18',
    wx.WXK_F19: 'F19',
    wx.WXK_F20: 'F20',
    wx.WXK_F21: 'F21',
    wx.WXK_F22: 'F22',
    wx.WXK_F23: 'F23',
    wx.WXK_F24: 'F24',
    wx.WXK_NUMLOCK: 'Num Lock',
    wx.WXK_SCROLL: 'Scroll Lock'
}

#-------------------------------------------------------------------------
#  Converts a keystroke event into a corresponding key name:
#-------------------------------------------------------------------------


def key_event_to_name(event):
    """ Converts a keystroke event into a corresponding key name.
    """
    key_code = event.GetKeyCode()
    if event.ControlDown() and (1 <= key_code <= 26):
        key = chr(key_code + 96)
    else:
        key = key_map.get(key_code)
        if key is None:
            try:
                key = chr(key_code)
            except:
                # Handle the case of strange keyboard codes (such as the Apple
                # keyboard 'apple' key):
                key = 'unknown'

    name = ''
    if event.AltDown():
        name = 'Alt'

    if event.ControlDown():
        name += '-Ctrl'

    if event.ShiftDown() and ((name != '') or (len(key) > 1)):
        name += '-Shift'

    if key == ' ':
        key = 'Space'

    if len(name) > 0:
        key = key.lower()

    name += ('-' + key)

    if name[:1] == '-':
        return name[1:]
    return name



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
from exception import NamingError, InvalidNameError, NameAlreadyBoundError
from exception import NameNotFoundError, NotContextError
from exception import OperationNotSupportedError

from address import Address
from binding import Binding
from context import Context
from context_adapter import ContextAdapter
from context_adapter_factory import ContextAdapterFactory
from dynamic_context import DynamicContext
from dir_context import DirContext
from initial_context import InitialContext
from initial_context_factory import InitialContextFactory
from naming_event import NamingEvent
from naming_manager import naming_manager
from object_factory import ObjectFactory
from object_serializer import ObjectSerializer
from py_context import PyContext
from py_object_factory import PyObjectFactory
from pyfs_context import PyFSContext
from pyfs_context_factory import PyFSContextFactory
from pyfs_initial_context_factory import PyFSInitialContextFactory
from pyfs_object_factory import PyFSObjectFactory
from pyfs_state_factory import PyFSStateFactory
from reference import Reference
from referenceable import Referenceable
from referenceable_state_factory import ReferenceableStateFactory
from state_factory import StateFactory



import os
import sys
import subprocess
import shlex


def which(program):
    """ find program on system environment path.
    
    From http://stackoverflow.com/questions/377017/
    """
    if sys.platform == "darwin":
        def is_exe(fpath):
            return (os.path.isdir(fpath) and fpath.endswith(".app")) or (os.path.isfile(fpath) and os.access(fpath, os.X_OK))
    else:
        def is_exe(fpath):
            return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def run_detach(program, args, fspath, replace_arg=None):
    # don't use posix so it will handle Windows backslash separators
    args = shlex.split(args, posix=False)
    found = False
    if replace_arg:
        new_args = []
        for a in args:
            if replace_arg in a:
                a.replace(replace_arg, fspath)
                found = True
            new_args.append(a)
        args = new_args
    if not found:
        args.append(fspath)

    args[0:0] = [program]
    program = which(program)
    if program is None:
        raise RuntimeError("%s not found on system path" % args[0])
    args[0] = program
    if sys.platform == "win32":
        p = subprocess.Popen(args, close_fds=True, creationflags=0x00000008|subprocess.CREATE_NEW_PROCESS_GROUP)
    elif sys.platform == "darwin":
        args[0:0] = ["open", "-a"]
        p = subprocess.call(args)
    else:
        p = subprocess.Popen(args, close_fds=True)
    return p



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought undo package component>
#------------------------------------------------------------------------------

from __future__ import absolute_import

# Enthought library imports.
from pyface.action.api import Action
from traits.api import Any, Callable, Instance
from ..i_command_stack import ICommandStack


class CommandAction(Action):
    """ The CommandAction class is an Action class that wraps undo/redo
    commands.  It is only useful for commands that do not take any arguments or
    return any result.
    """

    #### 'CommandAction' interface ############################################

    # The command to create when the action is performed.
    command = Callable

    # The command stack onto which the command will be pushed when the action
    # is performed.
    command_stack = Instance(ICommandStack)

    # This is the data on which the command operates.
    data = Any

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ This is reimplemented to push a new command instance onto the
        command stack.
        """

        self.command_stack.push(self.command(data=self.data))

    def _name_default(self):
        """ This gets the action name from the command. """

        if self.command:
            name = self.command().name
        else:
            name = ""

        return name



"""
fs.contrib.davfs
================


FS implementation accessing a WebDAV server.

This module provides a relatively-complete WebDAV Level 1 client that exposes
a WebDAV server as an FS object.  Locks are not currently supported.

Requires the dexml module:

    http://pypi.python.org/pypi/dexml/

"""
#  Copyright (c) 2009-2010, Cloud Matrix Pty. Ltd.
#  All rights reserved; available under the terms of the MIT License.

from __future__ import with_statement

import os
import sys
import httplib
import socket
from urlparse import urlparse
import stat as statinfo
from urllib import quote as urlquote
from urllib import unquote as urlunquote
import base64
import re
import time
import datetime
import cookielib
import fnmatch
import xml.dom.pulldom
import threading
from collections import deque

import fs
from fs.base import *
from fs.path import *
from fs.errors import *
from fs.remote import RemoteFileBuffer
from fs import iotools

from fs.contrib.davfs.util import *
from fs.contrib.davfs import xmlobj
from fs.contrib.davfs.xmlobj import *

import six
from six import b

import errno
_RETRYABLE_ERRORS = [errno.EADDRINUSE]
try:
    _RETRYABLE_ERRORS.append(errno.ECONNRESET)
    _RETRYABLE_ERRORS.append(errno.ECONNABORTED)
except AttributeError:
    _RETRYABLE_ERRORS.append(104)



class DAVFS(FS):
    """Access a remote filesystem via WebDAV.

    This FS implementation provides access to a remote filesystem via the
    WebDAV protocol.  Basic Level 1 WebDAV is supported; locking is not
    currently supported, but planned for the future.

    HTTP Basic authentication is supported; provide a dict giving username
    and password in the "credentials" argument, or a callback for obtaining
    one in the "get_credentials" argument.

    To use custom HTTP connector classes (e.g. to implement proper certificate
    checking for SSL connections) you can replace the factory functions in the
    DAVFS.connection_classes dictionary, or provide the "connection_classes"
    argument.
    """

    connection_classes = {
        "http":  httplib.HTTPConnection,
        "https":  httplib.HTTPSConnection,
    }

    _DEFAULT_PORT_NUMBERS = {
        "http": 80,
        "https": 443,
    }

    _meta = { 'virtual' : False,
              'read_only' : False,
              'unicode_paths' : True,
              'case_insensitive_paths' : False,
              'network' : True
             }

    def __init__(self,url,credentials=None,get_credentials=None,thread_synchronize=True,connection_classes=None,timeout=None):
        """DAVFS constructor.

        The only required argument is the root url of the remote server. If
        authentication is required, provide the 'credentials' keyword argument
        and/or the 'get_credentials' keyword argument.  The former is a dict
        of credentials info, while the latter is a callback function returning
        such a dict. Only HTTP Basic Auth is supported at this stage, so the
        only useful keys in a credentials dict are 'username' and 'password'.
        """
        if not url.endswith("/"):
            url = url + "/"
        self.url = url
        self.timeout = timeout
        self.credentials = credentials
        self.get_credentials = get_credentials
        if connection_classes is not None:
            self.connection_classes = self.connection_classes.copy()
            self.connection_classes.update(connection_classes)
        self._connections = []
        self._free_connections = {}
        self._connection_lock = threading.Lock()
        self._cookiejar = cookielib.CookieJar()
        super(DAVFS,self).__init__(thread_synchronize=thread_synchronize)
        #  Check that the server speaks WebDAV, and normalize the URL
        #  after any redirects have been followed.
        self.url = url
        pf = propfind(prop="<prop xmlns='DAV:'><resourcetype /></prop>")
        resp = self._request("/","PROPFIND",pf.render(),{"Depth":"0"})
        try:
            if resp.status == 404:
                raise ResourceNotFoundError("/",msg="root url gives 404")
            if resp.status in (401,403):
                raise PermissionDeniedError("listdir (http %s)" % resp.status)
            if resp.status != 207:
                msg = "server at %s doesn't speak WebDAV" % (self.url,)
                raise RemoteConnectionError("",msg=msg,details=resp.read())
        finally:
            resp.close()
        self.url = resp.request_url
        self._url_p = urlparse(self.url)

    def close(self):
        for con in self._connections:
            con.close()
        super(DAVFS,self).close()

    def _take_connection(self,url):
        """Get a connection to the given url's host, re-using if possible."""
        scheme = url.scheme.lower()
        hostname = url.hostname
        port = url.port
        if not port:
            try:
                port = self._DEFAULT_PORT_NUMBERS[scheme]
            except KeyError:
                msg = "unsupported protocol: '%s'" % (url.scheme,)
                raise RemoteConnectionError(msg=msg)
        #  Can we re-use an existing connection?
        with self._connection_lock:
            now = time.time()
            try:
                free_connections = self._free_connections[(hostname,port)]
            except KeyError:
                self._free_connections[(hostname,port)] = deque()
                free_connections = self._free_connections[(hostname,port)]
            else:
                while free_connections:
                    (when,con) = free_connections.popleft()
                    if when + 30 > now:
                        return (False,con)
                    self._discard_connection(con)
        #  Nope, we need to make a fresh one.
        try:
            ConClass = self.connection_classes[scheme]
        except KeyError:
            msg = "unsupported protocol: '%s'" % (url.scheme,)
            raise RemoteConnectionError(msg=msg)
        con = ConClass(url.hostname,url.port,timeout=self.timeout)
        self._connections.append(con)
        return (True,con)

    def _give_connection(self,url,con):
        """Return a connection to the pool, or destroy it if dead."""
        scheme = url.scheme.lower()
        hostname = url.hostname
        port = url.port
        if not port:
            try:
                port = self._DEFAULT_PORT_NUMBERS[scheme]
            except KeyError:
                msg = "unsupported protocol: '%s'" % (url.scheme,)
                raise RemoteConnectionError(msg=msg)
        with self._connection_lock:
            now = time.time()
            try:
                free_connections = self._free_connections[(hostname,port)]
            except KeyError:
                self._free_connections[(hostname,port)] = deque()
                free_connections = self._free_connections[(hostname,port)]
            free_connections.append((now,con))

    def _discard_connection(self,con):
        con.close()
        self._connections.remove(con)

    def __str__(self):
        return '<DAVFS: %s>' % (self.url,)
    __repr__ = __str__

    def __getstate__(self):
        state = super(DAVFS,self).__getstate__()
        del state["_connection_lock"]
        del state["_connections"]
        del state["_free_connections"]
        # Python2.5 cannot load pickled urlparse.ParseResult objects.
        del state["_url_p"]
        # CookieJar objects contain a lock, so they can't be pickled.
        del state["_cookiejar"]
        return state

    def __setstate__(self,state):
        super(DAVFS,self).__setstate__(state)
        self._connections = []
        self._free_connections = {}
        self._connection_lock = threading.Lock()
        self._url_p = urlparse(self.url)
        self._cookiejar = cookielib.CookieJar()

    def getpathurl(self, path, allow_none=False):
        """Convert a client-side path into a server-side URL."""
        path = relpath(normpath(path))
        if path.endswith("/"):
            path = path[:-1]
        if isinstance(path,unicode):
            path = path.encode("utf8")
        return self.url + urlquote(path)

    def _url2path(self,url):
        """Convert a server-side URL into a client-side path."""
        path = urlunquote(urlparse(url).path)
        root = urlunquote(self._url_p.path)
        path = path[len(root)-1:].decode("utf8")
        while path.endswith("/"):
            path = path[:-1]
        return path

    def _isurl(self,path,url):
        """Check whether the given URL corresponds to the given local path."""
        path = normpath(relpath(path))
        upath = relpath(normpath(self._url2path(url)))
        return path == upath

    def _request(self,path,method,body="",headers={}):
        """Issue a HTTP request to the remote server.

        This is a simple wrapper around httplib that does basic error and
        sanity checking e.g. following redirects and providing authentication.
        """
        url = self.getpathurl(path)
        visited = []
        resp = None
        try:
            resp = self._raw_request(url,method,body,headers)
            #  Loop to retry for redirects and authentication responses.
            while resp.status in (301,302,401,403):
                resp.close()
                if resp.status in (301,302,):
                    visited.append(url)
                    url = resp.getheader("Location",None)
                    if not url:
                        raise OperationFailedError(msg="no location header in 301 response")
                    if url in visited:
                        raise OperationFailedError(msg="redirection seems to be looping")
                    if len(visited) > 10:
                        raise OperationFailedError("too much redirection")
                elif resp.status in (401,403):
                    if self.get_credentials is None:
                        break
                    else:
                        creds = self.get_credentials(self.credentials)
                        if creds is None:
                            break
                        else:
                            self.credentials = creds
                resp = self._raw_request(url,method,body,headers)
        except Exception:
            if resp is not None:
                resp.close()
            raise
        resp.request_url = url
        return resp

    def _raw_request(self,url,method,body,headers,num_tries=0):
        """Perform a single HTTP request, without any error handling."""
        if self.closed:
            raise RemoteConnectionError("",msg="FS is closed")
        if isinstance(url,basestring):
            url = urlparse(url)
        if self.credentials is not None:
            username = self.credentials.get("username","")
            password = self.credentials.get("password","")
            if username is not None and password is not None:
                creds = "%s:%s" % (username,password,)
                creds = "Basic %s" % (base64.b64encode(creds).strip(),)
                headers["Authorization"] = creds
        (size,chunks) = normalize_req_body(body)
        try:
            (fresh,con) = self._take_connection(url)
            try:
                con.putrequest(method,url.path)
                if size is not None:
                    con.putheader("Content-Length",str(size))
                if hasattr(body,"md5"):
                    md5 = body.md5.decode("hex").encode("base64")
                    con.putheader("Content-MD5",md5)
                for hdr,val in headers.iteritems():
                    con.putheader(hdr,val)
                self._cookiejar.add_cookie_header(FakeReq(con,url.scheme,url.path))
                con.endheaders()
                for chunk in chunks:
                    con.send(chunk)
                    if self.closed:
                        raise RemoteConnectionError("",msg="FS is closed")
                resp = con.getresponse()
                self._cookiejar.extract_cookies(FakeResp(resp),FakeReq(con,url.scheme,url.path))
            except Exception:
                self._discard_connection(con)
                raise
            else:
                old_close = resp.close
                def new_close():
                    del resp.close
                    old_close()
                    con.close()
                    self._give_connection(url,con)
                resp.close = new_close
                return resp
        except socket.error, e:
            if not fresh:
                return self._raw_request(url,method,body,headers,num_tries)
            if e.args[0] in _RETRYABLE_ERRORS:
                if num_tries < 3:
                    num_tries += 1
                    return self._raw_request(url,method,body,headers,num_tries)
            try:
                msg = e.args[1]
            except IndexError:
                msg = str(e)
            raise RemoteConnectionError("",msg=msg,details=e)

    def setcontents(self,path, data=b'', encoding=None, errors=None, chunk_size=1024 * 64):
        if isinstance(data, six.text_type):
            data = data.encode(encoding=encoding, errors=errors)
        resp = self._request(path, "PUT", data)
        resp.close()
        if resp.status == 405:
            raise ResourceInvalidError(path)
        if resp.status == 409:
            raise ParentDirectoryMissingError(path)
        if resp.status not in (200,201,204):
            raise_generic_error(resp,"setcontents",path)

    @iotools.filelike_to_stream
    def open(self,path,mode="r", **kwargs):
        mode = mode.replace("b","").replace("t","")
        # Truncate the file if requested
        contents = b("")
        if "w" in mode:
            self.setcontents(path,contents)
        else:
            contents = self._request(path,"GET")
            if contents.status == 404:
                # Create the file if it's missing in append mode.
                if "a" not in mode:
                    contents.close()
                    raise ResourceNotFoundError(path)
                contents = b("")
                self.setcontents(path,contents)
            elif contents.status in (401,403):
                contents.close()
                raise PermissionDeniedError("open")
            elif contents.status != 200:
                contents.close()
                raise_generic_error(contents,"open",path)
            elif self.isdir(path):
                contents.close()
                raise ResourceInvalidError(path)
        #  For streaming reads, return the socket contents directly.
        if mode == "r-":
            contents.size = contents.getheader("Content-Length",None)
            if contents.size is not None:
                try:
                    contents.size = int(contents.size)
                except ValueError:
                    contents.size = None
            if not hasattr(contents,"__exit__"):
                contents.__enter__ = lambda *a: contents
                contents.__exit__ = lambda *a: contents.close()
            return contents
        #  For everything else, use a RemoteFileBuffer.
        #  This will take care of closing the socket when it's done.
        return RemoteFileBuffer(self,path,mode,contents)

    def exists(self,path):
        pf = propfind(prop="<prop xmlns='DAV:'><resourcetype /></prop>")
        response = self._request(path,"PROPFIND",pf.render(),{"Depth":"0"})
        response.close()
        if response.status == 207:
            return True
        if response.status == 404:
            return False
        raise_generic_error(response,"exists",path)

    def isdir(self,path):
        pf = propfind(prop="<prop xmlns='DAV:'><resourcetype /></prop>")
        response = self._request(path,"PROPFIND",pf.render(),{"Depth":"0"})
        try:
            if response.status == 404:
                return False
            if response.status != 207:
                raise_generic_error(response,"isdir",path)
            body = response.read()
            msres = multistatus.parse(body)
            for res in msres.responses:
                if self._isurl(path,res.href):
                   for ps in res.propstats:
                       if ps.props.getElementsByTagNameNS("DAV:","collection"):
                           return True
            return False
        finally:
            response.close()

    def isfile(self,path):
        pf = propfind(prop="<prop xmlns='DAV:'><resourcetype /></prop>")
        response = self._request(path,"PROPFIND",pf.render(),{"Depth":"0"})
        try:
            if response.status == 404:
                return False
            if response.status != 207:
                raise_generic_error(response,"isfile",path)
            msres = multistatus.parse(response.read())
            for res in msres.responses:
               if self._isurl(path,res.href):
                  for ps in res.propstats:
                     rt = ps.props.getElementsByTagNameNS("DAV:","resourcetype")
                     cl = ps.props.getElementsByTagNameNS("DAV:","collection")
                     if rt and not cl:
                        return True
            return False
        finally:
            response.close()

    def listdir(self,path="./",wildcard=None,full=False,absolute=False,dirs_only=False,files_only=False):
        return list(self.ilistdir(path=path,wildcard=wildcard,full=full,absolute=absolute,dirs_only=dirs_only,files_only=files_only))

    def ilistdir(self,path="./",wildcard=None,full=False,absolute=False,dirs_only=False,files_only=False):
        props = "<D:resourcetype />"
        dir_ok = False
        for res in self._do_propfind(path,props):
            if self._isurl(path,res.href):
               # The directory itself, check it's actually a directory
               for ps in res.propstats:
                   if ps.props.getElementsByTagNameNS("DAV:","collection"):
                      dir_ok = True
                      break
            else:
                nm = basename(self._url2path(res.href))
                entry_ok = False
                if dirs_only:
                    for ps in res.propstats:
                        if ps.props.getElementsByTagNameNS("DAV:","collection"):
                            entry_ok = True
                            break
                elif files_only:
                    for ps in res.propstats:
                        if ps.props.getElementsByTagNameNS("DAV:","collection"):
                            break
                    else:
                        entry_ok = True
                else:
                    entry_ok = True
                if not entry_ok:
                    continue
                if wildcard is not None:
                    if isinstance(wildcard,basestring):
                        if not fnmatch.fnmatch(nm,wildcard):
                            continue
                    else:
                        if not wildcard(nm):
                            continue
                if full:
                    yield relpath(pathjoin(path,nm))
                elif absolute:
                    yield abspath(pathjoin(path,nm))
                else:
                    yield nm
        if not dir_ok:
            raise ResourceInvalidError(path)

    def listdirinfo(self,path="./",wildcard=None,full=False,absolute=False,dirs_only=False,files_only=False):
        return list(self.ilistdirinfo(path=path,wildcard=wildcard,full=full,absolute=absolute,dirs_only=dirs_only,files_only=files_only))

    def ilistdirinfo(self,path="./",wildcard=None,full=False,absolute=False,dirs_only=False,files_only=False):
        props = "<D:resourcetype /><D:getcontentlength />" \
                "<D:getlastmodified /><D:getetag />"
        dir_ok = False
        for res in self._do_propfind(path,props):
            if self._isurl(path,res.href):
               # The directory itself, check it's actually a directory
               for ps in res.propstats:
                   if ps.props.getElementsByTagNameNS("DAV:","collection"):
                      dir_ok = True
                      break
            else:
                # An entry in the directory, check if it's of the
                # appropriate type and add to entries list as required.
                info = self._info_from_propfind(res)
                nm = basename(self._url2path(res.href))
                entry_ok = False
                if dirs_only:
                    for ps in res.propstats:
                        if ps.props.getElementsByTagNameNS("DAV:","collection"):
                            entry_ok = True
                            break
                elif files_only:
                    for ps in res.propstats:
                        if ps.props.getElementsByTagNameNS("DAV:","collection"):
                            break
                    else:
                        entry_ok = True
                else:
                    entry_ok = True
                if not entry_ok:
                    continue
                if wildcard is not None:
                    if isinstance(wildcard,basestring):
                        if not fnmatch.fnmatch(nm,wildcard):
                            continue
                    else:
                        if not wildcard(nm):
                            continue
                if full:
                    yield (relpath(pathjoin(path,nm)),info)
                elif absolute:
                    yield (abspath(pathjoin(path,nm)),info)
                else:
                    yield (nm,info)
        if not dir_ok:
            raise ResourceInvalidError(path)

    def makedir(self,path,recursive=False,allow_recreate=False):
        response = self._request(path,"MKCOL")
        response.close()
        if response.status == 201:
            return True
        if response.status == 409:
            if not recursive:
                raise ParentDirectoryMissingError(path)
            self.makedir(dirname(path),recursive=True,allow_recreate=True)
            self.makedir(path,recursive=False,allow_recreate=allow_recreate)
            return True
        if response.status == 405:
            if not self.isdir(path):
                raise ResourceInvalidError(path)
            if not allow_recreate:
                raise DestinationExistsError(path)
            return True
        if response.status < 200 or response.status >= 300:
            raise_generic_error(response,"makedir",path)

    def remove(self,path):
        if self.isdir(path):
            raise ResourceInvalidError(path)
        response = self._request(path,"DELETE")
        response.close()
        if response.status == 405:
            raise ResourceInvalidError(path)
        if response.status < 200 or response.status >= 300:
            raise_generic_error(response,"remove",path)
        return True

    def removedir(self,path,recursive=False,force=False):
        if self.isfile(path):
            raise ResourceInvalidError(path)
        if not force and self.listdir(path):
            raise DirectoryNotEmptyError(path)
        response = self._request(path,"DELETE")
        response.close()
        if response.status == 405:
            raise ResourceInvalidError(path)
        if response.status < 200 or response.status >= 300:
            raise_generic_error(response,"removedir",path)
        if recursive and path not in ("","/"):
            try:
                self.removedir(dirname(path),recursive=True)
            except DirectoryNotEmptyError:
                pass
        return True

    def rename(self,src,dst):
        self._move(src,dst)

    def getinfo(self,path):
        info = {}
        info["name"] = basename(path)
        pf = propfind(prop="<prop xmlns='DAV:'><resourcetype /><getcontentlength /><getlastmodified /><getetag /></prop>")
        response = self._request(path,"PROPFIND",pf.render(),{"Depth":"0"})
        try:
            if response.status != 207:
                raise_generic_error(response,"getinfo",path)
            msres = multistatus.parse(response.read())
            for res in msres.responses:
                if self._isurl(path,res.href):
                    info.update(self._info_from_propfind(res))
            if "st_mode" not in info:
               info["st_mode"] = 0700 | statinfo.S_IFREG
            return info
        finally:
            response.close()

    def _do_propfind(self,path,props):
        """Incremental PROPFIND parsing, for use with ilistdir/ilistdirinfo.

        This generator method incrementally parses the results returned by
        a PROPFIND, yielding each <response> object as it becomes available.
        If the server is able to send responses in chunked encoding, then
        this can substantially speed up iterating over the results.
        """
        pf = propfind(prop="<D:prop xmlns:D='DAV:'>"+props+"</D:prop>")
        response = self._request(path,"PROPFIND",pf.render(),{"Depth":"1"})
        try:
            if response.status == 404:
                raise ResourceNotFoundError(path)
            if response.status != 207:
                raise_generic_error(response,"listdir",path)
            xmlevents = xml.dom.pulldom.parse(response,bufsize=1024)
            for (evt,node) in xmlevents:
                if evt == xml.dom.pulldom.START_ELEMENT:
                    if node.namespaceURI == "DAV:":
                        if node.localName == "response":
                            xmlevents.expandNode(node)
                            yield xmlobj.response.parse(node)
        finally:
            response.close()

    def _info_from_propfind(self,res):
        info = {}
        for ps in res.propstats:
            findElements = ps.props.getElementsByTagNameNS
            # TODO: should check for status of the propfind first...
            # check for directory indicator
            if findElements("DAV:","collection"):
                info["st_mode"] = 0700 | statinfo.S_IFDIR
            # check for content length
            cl = findElements("DAV:","getcontentlength")
            if cl:
                cl = "".join(c.nodeValue for c in cl[0].childNodes)
                try:
                    info["size"] = int(cl)
                except ValueError:
                    pass
            # check for last modified time
            lm = findElements("DAV:","getlastmodified")
            if lm:
                lm = "".join(c.nodeValue for c in lm[0].childNodes)
                try:
                    # TODO: more robust datetime parsing
                    fmt = "%a, %d %b %Y %H:%M:%S GMT"
                    mtime = datetime.datetime.strptime(lm,fmt)
                    info["modified_time"] = mtime
                except ValueError:
                    pass
            # check for etag
            etag = findElements("DAV:","getetag")
            if etag:
                etag = "".join(c.nodeValue for c in etag[0].childNodes)
                if etag:
                    info["etag"] = etag
        if "st_mode" not in info:
            info["st_mode"] = 0700 | statinfo.S_IFREG
        return info


    def copy(self,src,dst,overwrite=False,chunk_size=None):
        if self.isdir(src):
            msg = "Source is not a file: %(path)s"
            raise ResourceInvalidError(src, msg=msg)
        self._copy(src,dst,overwrite=overwrite)

    def copydir(self,src,dst,overwrite=False,ignore_errors=False,chunk_size=0):
        if self.isfile(src):
            msg = "Source is not a directory: %(path)s"
            raise ResourceInvalidError(src, msg=msg)
        self._copy(src,dst,overwrite=overwrite)

    def _copy(self,src,dst,overwrite=False):
        headers = {"Destination":self.getpathurl(dst)}
        if overwrite:
            headers["Overwrite"] = "T"
        else:
            headers["Overwrite"] = "F"
        response = self._request(src,"COPY",headers=headers)
        response.close()
        if response.status == 412:
            raise DestinationExistsError(dst)
        if response.status == 409:
            raise ParentDirectoryMissingError(dst)
        if response.status < 200 or response.status >= 300:
            raise_generic_error(response,"copy",src)

    def move(self,src,dst,overwrite=False,chunk_size=None):
        if self.isdir(src):
            msg = "Source is not a file: %(path)s"
            raise ResourceInvalidError(src, msg=msg)
        self._move(src,dst,overwrite=overwrite)

    def movedir(self,src,dst,overwrite=False,ignore_errors=False,chunk_size=0):
        if self.isfile(src):
            msg = "Source is not a directory: %(path)s"
            raise ResourceInvalidError(src, msg=msg)
        self._move(src,dst,overwrite=overwrite)

    def _move(self,src,dst,overwrite=False):
        headers = {"Destination":self.getpathurl(dst)}
        if overwrite:
            headers["Overwrite"] = "T"
        else:
            headers["Overwrite"] = "F"
        response = self._request(src,"MOVE",headers=headers)
        response.close()
        if response.status == 412:
            raise DestinationExistsError(dst)
        if response.status == 409:
            raise ParentDirectoryMissingError(dst)
        if response.status < 200 or response.status >= 300:
            raise_generic_error(response,"move",src)

    @staticmethod
    def _split_xattr(name):
        """Split extended attribute name into (namespace,localName) pair."""
        idx = len(name)-1
        while idx >= 0 and name[idx].isalnum():
            idx -= 1
        return (name[:idx+1],name[idx+1:])

    def getxattr(self,path,name,default=None):
        (namespaceURI,localName) = self._split_xattr(name)
        # TODO: encode xml character entities in the namespace
        if namespaceURI:
            pf = propfind(prop="<D:prop xmlns:D='DAV:' xmlns='"+namespaceURI+"'><"+localName+" /></D:prop>")
        else:
            pf = propfind(prop="<D:prop xmlns:D='DAV:'><"+localName+" /></D:prop>")
        response = self._request(path,"PROPFIND",pf.render(),{"Depth":"0"})
        try:
            if response.status != 207:
                raise_generic_error(response,"getxattr",path)
            msres = multistatus.parse(response.read())
        finally:
            response.close()
        for res in msres.responses:
            if self._isurl(path,res.href):
               for ps in res.propstats:
                   if namespaceURI:
                       findElements = ps.props.getElementsByTagNameNS
                       propNode = findElements(namespaceURI,localName)
                   else:
                       findElements = ps.props.getElementsByTagName
                       propNode = findElements(localName)
                   if propNode:
                       propNode = propNode[0]
                       if ps.status.code == 200:
                         return "".join(c.toxml() for c in propNode.childNodes)
                       if ps.status.code == 404:
                         return default
                   raise OperationFailedError("getxattr",msres.render())
        return default

    def setxattr(self,path,name,value):
        (namespaceURI,localName) = self._split_xattr(name)
        # TODO: encode xml character entities in the namespace
        if namespaceURI:
            p = "<%s xmlns='%s'>%s</%s>" % (localName,namespaceURI,value,localName)
        else:
            p = "<%s>%s</%s>" % (localName,value,localName)
        pu = propertyupdate()
        pu.commands.append(set(props="<D:prop xmlns:D='DAV:'>"+p+"</D:prop>"))
        response = self._request(path,"PROPPATCH",pu.render(),{"Depth":"0"})
        response.close()
        if response.status < 200 or response.status >= 300:
            raise_generic_error(response,"setxattr",path)

    def delxattr(self,path,name):
        (namespaceURI,localName) = self._split_xattr(name)
        # TODO: encode xml character entities in the namespace
        if namespaceURI:
            p = "<%s xmlns='%s' />" % (localName,namespaceURI,)
        else:
            p = "<%s />" % (localName,)
        pu = propertyupdate()
        pu.commands.append(remove(props="<D:prop xmlns:D='DAV:'>"+p+"</D:prop>"))
        response = self._request(path,"PROPPATCH",pu.render(),{"Depth":"0"})
        response.close()
        if response.status < 200 or response.status >= 300:
            raise_generic_error(response,"delxattr",path)

    def listxattrs(self,path):
        pf = propfind(propname=True)
        response = self._request(path,"PROPFIND",pf.render(),{"Depth":"0"})
        try:
            if response.status != 207:
                raise_generic_error(response,"listxattrs",path)
            msres = multistatus.parse(response.read())
        finally:
            response.close()
        props = []
        for res in msres.responses:
            if self._isurl(path,res.href):
               for ps in res.propstats:
                   for node in ps.props.childNodes:
                       if node.nodeType != node.ELEMENT_NODE:
                           continue
                       if node.namespaceURI:
                           if node.namespaceURI in ("DAV:","PYFS:",):
                               continue
                           propname = node.namespaceURI + node.localName
                       else:
                           propname = node.nodeName
                       props.append(propname)
        return props

    # TODO: bulk getxattrs() and setxattrs() methods



def raise_generic_error(response,opname,path):
    if response.status == 404:
        raise ResourceNotFoundError(path,details=response.read())
    if response.status in (401,403):
        raise PermissionDeniedError(opname,details=response.read())
    if response.status == 423:
        raise ResourceLockedError(path,opname=opname,details=response.read())
    if response.status == 501:
        raise UnsupportedError(opname,details=response.read())
    if response.status == 405:
        raise ResourceInvalidError(path,opname=opname,details=response.read())
    raise OperationFailedError(opname,msg="Server Error: %s" % (response.status,),details=response.read())




#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   03/03/2006
#
#------------------------------------------------------------------------------

""" Defines the BasicUIEditor class, which allows creating editors that define
    their function by creating an embedded Traits UI.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from traitsui.ui_editor \
    import UIEditor as BaseUIEditor

from editor \
    import Editor

#-------------------------------------------------------------------------
#  'UIEditor' base class:
#-------------------------------------------------------------------------


class UIEditor(BaseUIEditor, Editor):
    """ An editor that creates an embedded Traits UI.
    """
    pass

#-- End UI preference save/restore interface -----------------------------



"""

  fs.tests.test_watch:  testcases for change watcher support

"""

import os
import sys
import time
import gc
import pickle
import unittest

from fs.path import *
from fs.errors import *
from fs.watch import *
from fs.tests import FSTestCases

try:
    from fs.osfs import watch_inotify
except ImportError:
    watch_inotify = None

if sys.platform == "win32":
    try:
        from fs.osfs import watch_win32
    except ImportError:
        watch_win32 = None
else:
    watch_win32 = None

import logging
logging.getLogger('pyinotify').setLevel(logging.ERROR)


import six
from six import PY3, b

class WatcherTestCases:
    """Testcases for filesystems providing change watcher support.

    This class should be used as a mixin to the unittest.TestCase class
    for filesystems that provide change watcher support.
    """

    def setupWatchers(self):
        self._captured_events = []
        self.watchfs.add_watcher(self._captured_events.append)

    def clearCapturedEvents(self):
        del self._captured_events[:]

    def waitForEvents(self):
        if isinstance(self.watchfs,PollingWatchableFS):
            self.watchfs._poll_cond.acquire()
            self.watchfs._poll_cond.wait()
            self.watchfs._poll_cond.wait()
            self.watchfs._poll_cond.release()
        else:
            time.sleep(2)

    def assertEventOccurred(self,cls,path=None,event_list=None,**attrs):
        if not self.checkEventOccurred(cls,path,event_list,**attrs):
            args = (cls.__name__,path,attrs)
            assert False, "Event did not occur: %s(%s,%s)" % args

    def checkEventOccurred(self,cls,path=None,event_list=None,**attrs):
        if event_list is None:
            event_list = self._captured_events
        self.waitForEvents()
        for event in event_list:
            if isinstance(event,cls):
                if path is None or event.path == path:
                    for (k,v) in attrs.iteritems():
                        if getattr(event,k) != v:
                            break
                    else:
                        # all attrs match - found it!
                        return True
        return False

    def test_watch_makedir(self):
        self.setupWatchers()
        self.fs.makedir("test1")
        self.assertEventOccurred(CREATED,"/test1")

    def test_watch_makedir_with_two_watchers(self):
        self.setupWatchers()
        events2 = []
        self.watchfs.add_watcher(events2.append)
        self.fs.makedir("test1")
        self.assertEventOccurred(CREATED,"/test1")
        self.assertEventOccurred(CREATED,"/test1",event_list=events2)

    def test_watch_readfile(self):
        self.setupWatchers()
        self.fs.setcontents("hello", b("hello world"))
        self.assertEventOccurred(CREATED,"/hello")
        self.clearCapturedEvents()
        old_atime = self.fs.getinfo("hello").get("accessed_time")
        self.assertEquals(self.fs.getcontents("hello"), b("hello world"))
        if not isinstance(self.watchfs,PollingWatchableFS):
            #  Help it along by updting the atime.
            #  TODO: why is this necessary?
            if self.fs.hassyspath("hello"):
                syspath = self.fs.getsyspath("hello")
                mtime = os.stat(syspath).st_mtime
                atime = int(time.time())
                os.utime(self.fs.getsyspath("hello"),(atime,mtime))
            self.assertEventOccurred(ACCESSED,"/hello")
        elif old_atime is not None:
            #  Some filesystems don't update atime synchronously, or only
            #  update it if it's too old, or don't update it at all!
            #  Try to force the issue, wait for it to change, but eventually
            #  give up and bail out.
            for i in xrange(10):
                if self.fs.getinfo("hello").get("accessed_time") != old_atime:
                    if not self.checkEventOccurred(MODIFIED,"/hello"):
                        self.assertEventOccurred(ACCESSED,"/hello")
                    break
                time.sleep(0.2)
                if self.fs.hassyspath("hello"):
                    syspath = self.fs.getsyspath("hello")
                    mtime = os.stat(syspath).st_mtime
                    atime = int(time.time())
                    os.utime(self.fs.getsyspath("hello"),(atime,mtime))

    def test_watch_writefile(self):
        self.setupWatchers()
        self.fs.setcontents("hello", b("hello world"))
        self.assertEventOccurred(CREATED,"/hello")
        self.clearCapturedEvents()
        self.fs.setcontents("hello", b("hello again world"))
        self.assertEventOccurred(MODIFIED,"/hello")

    def test_watch_single_file(self):
        self.fs.setcontents("hello", b("hello world"))
        events = []
        self.watchfs.add_watcher(events.append,"/hello",(MODIFIED,))
        self.fs.setcontents("hello", b("hello again world"))
        self.fs.remove("hello")
        self.waitForEvents()
        for evt in events:
            assert isinstance(evt,MODIFIED)
            self.assertEquals(evt.path,"/hello")

    def test_watch_single_file_remove(self):
        self.fs.makedir("testing")
        self.fs.setcontents("testing/hello", b("hello world"))
        events = []
        self.watchfs.add_watcher(events.append,"/testing/hello",(REMOVED,))
        self.fs.setcontents("testing/hello", b("hello again world"))
        self.waitForEvents()
        self.fs.remove("testing/hello")
        self.waitForEvents()
        self.assertEquals(len(events),1)
        assert isinstance(events[0],REMOVED)
        self.assertEquals(events[0].path,"/testing/hello")

    def test_watch_iter_changes(self):
        changes = iter_changes(self.watchfs)
        self.fs.makedir("test1")
        self.fs.setcontents("test1/hello", b("hello world"))
        self.waitForEvents()
        self.fs.removedir("test1",force=True)
        self.waitForEvents()
        self.watchfs.close()
        #  Locate the CREATED(test1) event
        event = changes.next(timeout=1)
        while not isinstance(event,CREATED) or event.path != "/test1":
            event = changes.next(timeout=1)
        #  Locate the CREATED(test1/hello) event
        event = changes.next(timeout=1)
        while not isinstance(event,CREATED) or event.path != "/test1/hello":
            event = changes.next(timeout=1)
        #  Locate the REMOVED(test1) event
        event = changes.next(timeout=1)
        while not isinstance(event,REMOVED) or event.path != "/test1":
            event = changes.next(timeout=1)
        #  Locate the CLOSED event
        event = changes.next(timeout=1)
        while not isinstance(event,CLOSED):
            event = changes.next(timeout=1)
        #  That should be the last event in the list
        self.assertRaises(StopIteration,getattr(changes, "next"),timeout=1)
        changes.close()



from fs import tempfs, osfs
class TestWatchers_TempFS(unittest.TestCase,FSTestCases,WatcherTestCases):

    def setUp(self):
        self.fs = tempfs.TempFS()
        watchfs = osfs.OSFS(self.fs.root_path)
        self.watchfs = ensure_watchable(watchfs,poll_interval=0.1)
        if watch_inotify is not None:
            self.assertEquals(watchfs,self.watchfs)
        if watch_win32 is not None:
            self.assertEquals(watchfs,self.watchfs)

    def tearDown(self):
        self.watchfs.close()
        self.fs.close()

    def check(self, p):
        return self.fs.exists(p)


from fs import memoryfs
class TestWatchers_MemoryFS(unittest.TestCase,FSTestCases,WatcherTestCases):

    def setUp(self):
        self.fs = self.watchfs = WatchableFS(memoryfs.MemoryFS())

    def tearDown(self):
        self.watchfs.close()
        self.fs.close()

    def check(self, p):
        return self.fs.exists(p)


class TestWatchers_MemoryFS_polling(TestWatchers_MemoryFS):

    def setUp(self):
        self.fs = memoryfs.MemoryFS()
        self.watchfs = ensure_watchable(self.fs,poll_interval=0.1)



from contextlib import contextmanager

from .file import H5File


@contextmanager
def open_h5file(filename, mode='r+', **kwargs):
    """Context manager for reading an HDF5 file as an H5File object.

    Parameters
    ----------
    filename : str
        HDF5 file name.
    mode : str
        Mode to open the file:

        'r' : Read-only
        'w' : Write; create new file (an existing file would be deleted).
        'a' : Read and write to file; create if not existing
        'r+': Read and write to file; must already exist

    See `H5File` for additional keyword arguments.
    """
    h5 = H5File(filename, mode=mode, **kwargs)
    try:
        yield h5
    finally:
        h5.close()



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The interface for an interactive Python shell. """

# Enthought library imports.
from traits.api import Event

# Local imports.
from pyface.key_pressed_event import KeyPressedEvent
from pyface.i_widget import IWidget


class IPythonShell(IWidget):
    """ The interface for an interactive Python shell. """

    #### 'IPythonShell' interface #############################################

    #: A command has been executed.
    command_executed = Event

    #: A key has been pressed.
    key_pressed = Event(KeyPressedEvent)

    ###########################################################################
    # 'IPythonShell' interface.
    ###########################################################################

    def interpreter(self):
        """ Get the shell's interpreter

        Returns
        -------
        interpreter : InteractiveInterpreter instance
            Returns the InteractiveInterpreter instance.
        """

    def bind(self, name, value):
        """ Binds a name to a value in the interpreter's namespace.

        Parameters
        ----------
        name : str
            The python idetifier to bind the value to.
        value : any
            The python object to be bound into the interpreter's namespace.
        """

    def execute_command(self, command, hidden=True):
        """ Execute a command in the interpreter.

        Parameters
        ----------
        command : str
            A Python command to execute.
        hidden : bool
            If 'hidden' is True then nothing is shown in the shell - not even
            a blank line.
        """

    def execute_file(self, path, hidden=True):
        """ Execute a file in the interpeter.

        Parameters
        ----------
        path : str
            The path to the Python file to execute.
        hidden : bool
            If 'hidden' is True then nothing is shown in the shell - not even
            a blank line.
        """


class MPythonShell(object):
    """ The mixin class that contains common code for toolkit specific
    implementations of the IPythonShell interface.

    Implements: bind(), _on_command_executed()
    """

    ###########################################################################
    # 'IPythonShell' interface.
    ###########################################################################

    def bind(self, name, value):
        """ Binds a name to a value in the interpreter's namespace.

        Parameters
        ----------
        name : str
            The python idetifier to bind the value to.
        value : any
            The python object to be bound into the interpreter's namespace.
        """
        self.interpreter().locals[name] = value

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _on_command_executed(self):
        """ Called when a command has been executed in the shell. """

        self.command_executed = self
